@@ -3,6 +3,7 @@
  * \section{Higher-level IO functions}
  *
  * \subsection{Regular expressions}
+ * \index{regular expressions}
  *
  * Many of the higher-level functions use regular expressions.
  * Regular expressions are defined by strings with syntax nearly identical
@@ -84,12 +85,12 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2004 Mojave Group, Caltech
+ * Copyright (C) 2004-2006 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -98,16 +99,21 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{nogin@cs.caltech.edu}
  * @end[license]
  *)
 open Lm_debug
 open Lm_printf
 open Lm_parser
 open Lm_location
+open Lm_symbol
 
 open Omake_ir
 open Omake_env
@@ -136,7 +142,7 @@ let debug_parsing =
  * Concatenate files into a string.
  *
  * \begin{doc}
- * \subsection{cat}
+ * \fun{cat}
  *
  * \begin{verbatim}
  *     cat(files) : Sequence
@@ -183,7 +189,7 @@ let cat venv pos loc args =
  * Grep takes some flags.
  *
  * \begin{doc}
- * \subsection{grep}
+ * \fun{grep}
  *
  * \begin{verbatim}
  *    grep(pattern) : String  # input from stdin, default options
@@ -204,7 +210,11 @@ let cat venv pos loc args =
  * The options are:
  * \begin{description}
  * \item[q] If specified, the output from \verb+grep+ is not displayed.
- * \item[n] If specified, output lines include the filename.
+ * \item[h] If specified, output lines will not include the filename (default, when only one input
+ *          file is given).
+ * \item[n] If specified, output lines include the filename (default, when more than one input file
+ *          is given).
+ * \item[v] If specified, search for lines without a match instead of lines with a match,
  * \end{description}
  *
  * The \verb+pattern+ is a regular expression.
@@ -213,8 +223,11 @@ let cat venv pos loc args =
  * Otherwise, it returns \verb+false+.
  * \end{doc}
  *)
-let grep_quiet_flag = 1
-let grep_print_flag = 2
+type grep_flag =
+   GrepQuiet
+ | GrepPrint
+ | GrepNoPrint
+ | GrepNoMatch
 
 let grep_flags pos loc s =
    let len = String.length s in
@@ -224,16 +237,20 @@ let grep_flags pos loc s =
       else
          let flag =
             match s.[i] with
-               'q' | 'Q' ->
-                  grep_quiet_flag
+               'q' ->
+                  GrepQuiet
              | 'n' ->
-                  grep_print_flag
+                  GrepPrint
+             | 'v' ->
+                  GrepNoMatch
+             | 'h' ->
+                  GrepNoPrint
              | c ->
                   raise (OmakeException (loc_pos loc pos, StringStringError ("illegal grep option", String.make 1 c)))
          in
-            collect (flags lor flag) (succ i)
+            collect (flag::flags) (succ i)
    in
-      collect 0 0
+      collect [] 0
 
 let grep venv pos loc args =
    let pos = string_pos "grep" pos in
@@ -265,16 +282,19 @@ let grep venv pos loc args =
             flags, [venv_find_var venv ScopeGlobal pos loc stdin_sym]
        | [_] ->
             flags, files
-       | _ ->
-            flags lor grep_print_flag, files
+       | _::_::_ ->
+            (if List.mem GrepNoPrint flags then flags else GrepPrint :: flags), files
    in
+   let verbose = not (List.mem GrepQuiet flags) in
+   let print = List.mem GrepPrint flags in
+   let matches = not (List.mem GrepNoMatch flags) in
 
    (* Grep against a single line *)
    let grep_line file found line =
-      let b = lexer_matches pattern line in
-         if b && (flags land grep_quiet_flag) = 0 then
+      let b = ((lexer_matches pattern line) == matches) in
+         if b && verbose then
             begin
-               if (flags land grep_print_flag) <> 0 then
+               if print then
                   begin
                      Lm_channel.output_string outx file;
                      Lm_channel.output_char outx ':'
@@ -309,10 +329,7 @@ let grep venv pos loc args =
    in
    let b = List.fold_left grep_file false files in
       Lm_channel.flush outx;
-      if b then
-         val_true
-      else
-         val_false
+      val_of_bool b
 
 let builtin_grep venv pos loc args =
    let pos = string_pos "builtin-grep" pos in
@@ -334,6 +351,10 @@ let builtin_grep venv pos loc args =
                       collect ("q" ^ flags) args
                  | "-n" ->
                       collect ("n" ^ flags) args
+                 | "-v" ->
+                      collect ("v" ^ flags) args
+                 | "-h" ->
+                      collect ("h" ^ flags) args
                  | pattern ->
                       flags, pattern, args)
           | [] ->
@@ -345,7 +366,260 @@ let builtin_grep venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{awk}
+ * \fun{scan}
+ *
+ * \begin{verbatim}
+ *    scan(input-files)
+ *    case string1
+ *       body1
+ *    case string2
+ *       body2
+ *    ...
+ *    default
+ *       bodyd
+ * \end{verbatim}
+ *
+ * The \verb+scan+ function provides input processing in command-line form.
+ * The function takes file/filename arguments.  If called with no
+ * arguments, the input is taken from \verb+stdin+. If arguments are provided,
+ * each specifies an \verb+InChannel+, or the name of a file for input.
+ * Output is always to \verb+stdout+.
+ *
+ * The \verb+scan+ function operates by reading the input one line at a time,
+ * and processing it according to the following algorithm.
+ *
+ * For each line,
+ * the record is first split into fields, and
+ * the fields are bound to the variables \verb+$1, $2, ...+.  The variable
+ * \verb+$0+ is defined to be the entire line, and \verb+$*+ is an array
+ * of all the field values.  The \verb+$(NF)+ variable is defined to be the number
+ * of fields.
+ *
+ * Next, a case expression is selected.  If \verb+string_i+ matches the token \verb+$1+,
+ * then \verb+body_i+ is evaluated.  If the body ends in an \verb+export+, the state
+ * is passed to the next clause.  Otherwise the value is discarded.
+ *
+ * For example, here is an \verb+scan+ function that acts as a simple command processor.
+ *
+ * \begin{verbatim}
+ *     calc() =
+ *        i = 0
+ *        scan(script.in)
+ *        case print
+ *           println($i)
+ *        case inc
+ *           i = $(add $i, 1)
+ *           export
+ *        case dec
+ *           i = $(sub $i, 1)
+ *           export
+ *        case addconst
+ *           i = $(add $i, $2)
+ *           export
+ *        default
+ *           eprintln($"Unknown command: $1")
+ * \end{verbatim}
+ *
+ * The \verb+scan+ function also supports several options.
+ *
+ * \begin{verbatim}
+ *     scan(options, files)
+ *     ...
+ * \end{verbatim}
+ *
+ * \begin{description}
+ * \item[A] Parse each line as an argument list, where arguments
+ *    may be quoted.  For example, the following line has three words,
+ *    ``\verb+ls+'', ``\verb+-l+'', ``\verb+Program Files+''.
+ *
+ *    \begin{verbatim}
+ *        ls -l "Program Files"
+ *    \end{verbatim}
+ * \item[O] Parse each line using white space as the separator, using the
+ *    usual \OMake{} algorithm for string parsing.  This is the default.
+ * \item[x] Once each line is split, reduce each word using the
+ *    hex representation.  This is the usual hex representation used
+ *    in URL specifiers, so the string ``Program Files'' may be
+ *    alternately represented in the form Program%20Files or
+ *    Program+Files.
+ * \end{description}
+ *
+ * Note, if you want to redirect the output to a file, the easiest way is to
+ * redefine the \verb+stdout+ variable.  The \verb+stdout+ variable is scoped the
+ * same way as other variables, so this definition does not affect the meaning of
+ * \verb+stdout+ outside the \verb+calc+ function.
+ *
+ * \begin{verbatim}
+ *     calc() =
+ *         stdout = $(fopen script.out, w)
+ *         scan(script.in)
+ *            ...
+ *         close(stdout)
+ * \end{verbatim}
+ * \end{doc}
+ *)
+
+(*
+ * Scanner options.
+ *)
+type parse_option =
+   ParseArgs
+ | ParseWords
+
+type rewrite_option =
+   RewriteHex
+ | RewriteNone
+
+let scan_options venv pos loc options s =
+   let len = String.length s in
+   let rec collect ((poption, roption) as options) i =
+      if i = len then
+         options
+      else
+         let options =
+            match s.[i] with
+               'A' ->
+                  ParseArgs, roption
+             | 'O' ->
+                  ParseWords, roption
+             | 'x' ->
+                  poption, RewriteHex
+             | c ->
+                  raise (OmakeException (loc_pos loc pos, StringStringError ("illegal option", s)))
+         in
+            collect options (succ i)
+   in
+      collect options 0
+
+let scan_options venv pos loc options =
+   List.fold_left (scan_options venv pos loc) (ParseWords, RewriteNone) (strings_of_value venv pos options)
+
+(*
+ * The arguments.
+ *)
+let scan_args venv pos loc args =
+   let pos = string_pos "scan_args" pos in
+   let cases, options, files =
+      match args with
+         [ValCases cases] ->
+            cases, ValNone, venv_find_var venv ScopeGlobal pos loc stdin_sym
+       | [ValCases cases; files] ->
+            cases, ValNone, files
+       | [ValCases cases; options; files] ->
+            cases, options, files
+       | _ ->
+            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))
+   in
+   let poptions, roptions = scan_options venv pos loc options in
+      cases, poptions, roptions, values_of_value venv pos files
+
+(*
+ * Awk the value.
+ *)
+let scan venv pos loc args =
+   let pos = string_pos "scan" pos in
+   let cases, token_mode, rewrite_mode, files = scan_args venv pos loc args in
+
+   (* Get lexers for all the cases *)
+   let cases, def =
+      List.fold_left (fun (cases, def) (v, test, body) ->
+            if Lm_symbol.eq v case_sym then
+               let s = string_of_value venv pos test in
+               let cases =
+                  SymbolTable.filter_add cases (Lm_symbol.add s) (fun b ->
+                        match b with
+                           Some _ ->
+                              raise (OmakeException (loc_pos loc pos, StringVarError ("duplicate case", v)))
+                         | None ->
+                              body)
+               in
+                  cases, def
+            else if Lm_symbol.eq v default_sym then
+               match def with
+                  Some _ ->
+                     raise (OmakeException (loc_pos loc pos, StringError "duplicate default case"))
+                | None ->
+                     cases, Some body
+            else
+               raise (OmakeException (loc_pos loc pos, StringVarError ("unknown case", v)))) (SymbolTable.empty, None) cases
+   in
+
+   (* Split a line into words *)
+   let collect_words_argv line =
+      let words =
+         match token_mode with
+            ParseArgs ->
+               Lm_string_util.parse_args line
+          | ParseWords ->
+               strings_of_value venv pos (ValString line)
+      in
+         match rewrite_mode with
+            RewriteHex ->
+               List.map Lm_string_util.decode_hex_name words
+          | RewriteNone ->
+               words
+   in
+
+   (* Select a case and run it *)
+   let eval_case venv words =
+      let body =
+         match words with
+            command :: _ ->
+               (try Some (SymbolTable.find cases (Lm_symbol.add command)) with
+                   Not_found ->
+                      def)
+          | [] ->
+               def
+      in
+      let result =
+         match body with
+            Some body ->
+               eval_body_value venv pos body
+          | None ->
+               ValNone
+      in
+      let venv, _ = add_exports venv pos result in
+         venv
+   in
+
+   (* Read the file a line at a time *)
+   let rec line_loop venv inx =
+      let text =
+         try Some (Lm_channel.input_line inx) with
+            End_of_file ->
+               None
+      in
+         match text with
+            Some line ->
+               let words = collect_words_argv line in
+               let venv = venv_add_match venv line words in
+               let venv = eval_case venv words in
+                  line_loop venv inx
+          | None ->
+               venv
+   in
+   let rec file_loop venv args =
+      match args with
+         arg :: args ->
+            let inp, close_in = in_channel_of_any_value venv pos arg in
+            let inx = venv_find_channel venv pos inp in
+            let venv =
+               try line_loop venv inx with
+                  exn when close_in ->
+                     venv_close_channel venv pos inp;
+                     raise exn
+               in
+               if close_in then
+                  venv_close_channel venv pos inp;
+               file_loop venv args
+       | [] ->
+            venv
+   in
+      ValEnv (file_loop venv files, ExportAll)
+
+(*
+ * \begin{doc}
+ * \fun{awk}
  *
  * \begin{verbatim}
  *    awk(input-files)
@@ -356,12 +630,25 @@ let builtin_grep venv pos loc args =
  *    ...
  *    default:
  *       bodyd
+ *\end{verbatim}
+ *
+ * or
+ *
+ * \begin{verbatim}
+ *    awk(options, input-files)
+ *    case pattern1:
+ *       body1
+ *    case pattern2:
+ *       body2
+ *    ...
+ *    default:
+ *       bodyd
  * \end{verbatim}
  *
  * The \verb+awk+ function provides input processing similar to \Cmd{awk}{1},
- * but more limited. The function takes filename arguments.  If called with no
- * arguments, the input is taken from \verb+stdin+. If arguments are provided,
+ * but more limited.  The \verb+input-files+ argument is a sequence of values,
  * each specifies an \verb+InChannel+, or the name of a file for input.
+ * If called with no options and no file arguments, the input is taken from \verb+stdin+.
  * Output is always to \verb+stdout+.
  *
  * The variables \verb+RS+ and \verb+FS+ define record and field separators
@@ -420,30 +707,44 @@ let builtin_grep venv pos loc args =
  *            ...
  *         close(stdout)
  * \end{verbatim}
+ *
+ * Options.
+ * \begin{description}
+ * \item[b] ``Break'' when evaluating cases.  Only the first case that matches will be selected.
+ * \end{description}
+ *
+ * The \hyperfun{break} can be used to abort the loop,
+ * exiting the \verb+awk+ function immediately.
  * \end{doc}
  *)
 
 (*
  * Evaluate all the cases that match.
  *)
-let rec awk_eval_cases venv pos loc line cases =
+let rec awk_eval_cases venv pos loc break line cases =
    match cases with
       (None, body) :: cases ->
          let result = eval_body_value venv pos body in
          let venv, _ = add_exports venv pos result in
-            awk_eval_cases venv pos loc line cases
+            if break then
+               venv
+            else
+               awk_eval_cases venv pos loc break line cases
     | (Some lex, body) :: cases ->
          let channel = Lm_channel.of_string line in
-         let result =
+         let result, stop =
             match Lexer.search lex channel with
                Some (_, _, _, _, args) ->
                   let venv = venv_add_match_args venv args in
-                     eval_body_value venv pos body
+                     eval_body_value venv pos body, break
              | None ->
-                  ValNone
+                  ValNone, false
          in
          let venv, _ = add_exports venv pos result in
-            awk_eval_cases venv pos loc line cases
+            if stop then
+               venv
+            else
+               awk_eval_cases venv pos loc break line cases
     | [] ->
          venv
 
@@ -452,34 +753,54 @@ let rec awk_eval_cases venv pos loc line cases =
  *)
 let awk_args venv pos loc args =
    let pos = string_pos "awk_args" pos in
-   let cases, files =
       match args with
          [ValCases cases] ->
             cases, [venv_find_var venv ScopeGlobal pos loc stdin_sym]
-       | ValCases cases :: files ->
-            cases, files
+       | [ValCases cases; files] ->
+            cases, values_of_value venv pos files
        | _ ->
-            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))
-   in
-      cases, files
+            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityRange (1, 2), List.length args)))
 
-let awk_args venv pos loc args =
+let awk_option_args venv pos loc args =
    let pos = string_pos "awk_args" pos in
-   let rec flatten args' args =
       match args with
-         arg :: args ->
-            flatten (List.rev_append (values_of_value venv pos arg) args') args
-       | [] ->
-            List.rev args'
+         [ValCases cases] ->
+            cases, "", [venv_find_var venv ScopeGlobal pos loc stdin_sym]
+       | [ValCases cases; files] ->
+            cases, "", values_of_value venv pos files
+       | [ValCases cases; options; files] ->
+            cases, string_of_value venv pos options, values_of_value venv pos files
+       | _ ->
+            raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityRange (1, 3), List.length args)))
+
+type awk_flag =
+   AwkBreak
+
+let awk_flags pos loc s =
+   let len = String.length s in
+   let rec collect flags i =
+      if i = len then
+         flags
+      else
+         let flag =
+            match s.[i] with
+               'b' ->
+                  AwkBreak
+             | c ->
+                  raise (OmakeException (loc_pos loc pos, StringStringError ("illegal awk option", String.make 1 c)))
+         in
+            collect (flag :: flags) (succ i)
    in
-      awk_args venv pos loc (flatten [] args)
+      collect [] 0
 
 (*
  * Awk the value.
  *)
 let awk venv pos loc args =
    let pos = string_pos "awk" pos in
-   let cases, files = awk_args venv pos loc args in
+   let cases, flags, files = awk_option_args venv pos loc args in
+   let flags = awk_flags pos loc flags in
+   let break = List.mem AwkBreak flags in
 
    (* Separator expressions *)
    let rs =
@@ -493,16 +814,16 @@ let awk venv pos loc args =
             "[ \t]+"
    in
    let rs_lex =
-      try lexer_of_string rs
-      with Failure err ->
-         let msg = sprintf "Mailformed regular expression '%s'" rs in
-            raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+      try lexer_of_string rs with
+         Failure err ->
+            let msg = sprintf "Malformed regular expression '%s'" rs in
+               raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
    in
    let fs_lex =
-      try lexer_of_string fs
-      with Failure err ->
-         let msg = sprintf "Mailformed regular expression '%s'" fs in
-            raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+      try lexer_of_string fs with
+         Failure err ->
+            let msg = sprintf "Malformed regular expression '%s'" fs in
+               raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
    in
 
    (* Get lexers for all the cases *)
@@ -511,10 +832,10 @@ let awk venv pos loc args =
             if Lm_symbol.eq v case_sym then
                let s = string_of_value venv pos test in
                let _, lex =
-                  try Lexer.add_clause Lexer.empty v s
-                  with Failure err ->
-                     let msg = sprintf "Mailformed regular expression '%s'" s in
-                        raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+                  try Lexer.add_clause Lexer.empty v s with
+                     Failure err ->
+                        let msg = sprintf "Malformed regular expression '%s'" s in
+                           raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
                in
                   Some lex, body
             else if Lm_symbol.eq v default_sym then
@@ -547,7 +868,7 @@ let awk venv pos loc args =
             (* Split into words *)
             let words = collect_words line in
             let venv = venv_add_match venv line words in
-            let venv = awk_eval_cases venv pos loc line cases in
+            let venv = awk_eval_cases venv pos loc break line cases in
                line_loop venv inx
    in
    let rec file_loop venv args =
@@ -555,18 +876,28 @@ let awk venv pos loc args =
          arg :: args ->
             let inp, close_in = in_channel_of_any_value venv pos arg in
             let inx = venv_find_channel venv pos inp in
-            let venv = line_loop venv inx in
+            let venv =
+               try line_loop venv inx with
+                  exn when close_in ->
+                     venv_close_channel venv pos inp;
+                     raise exn
+            in
                if close_in then
                   venv_close_channel venv pos inp;
                file_loop venv args
        | [] ->
             venv
    in
-      ValEnv (file_loop venv files, ExportAll)
+   let venv =
+      try file_loop venv files with
+         Break (_, venv) ->
+            venv
+   in
+      ValEnv (venv, ExportAll)
 
 (*
  * \begin{doc}
- * \subsection{fsubst}
+ * \fun{fsubst}
  *
  * \begin{verbatim}
  *    fsubst(files)
@@ -676,10 +1007,10 @@ let fsubst venv pos loc args =
             "\r|\n|\r\n"
    in
    let rs_lex =
-      try lexer_of_string rs
-      with Failure err ->
-         let msg = sprintf "Mailformed regular expression '%s'" rs in
-            raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+      try lexer_of_string rs with
+         Failure err ->
+            let msg = sprintf "Malformed regular expression '%s'" rs in
+               raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
    in
 
    (* Get lexers for all the cases *)
@@ -706,10 +1037,10 @@ let fsubst venv pos loc args =
                      subst_options venv pos loc options (string_of_value venv pos arg)) 0 options
             in
             let _, lex =
-               try Lexer.add_clause Lexer.empty v pattern
-               with Failure err ->
-                  let msg = sprintf "Mailformed regular expression '%s'" pattern in
-                     raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+               try Lexer.add_clause Lexer.empty v pattern with
+                  Failure err ->
+                     let msg = sprintf "Malformed regular expression '%s'" pattern in
+                        raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
             in
                lex, options, body) cases
    in
@@ -733,20 +1064,303 @@ let fsubst venv pos loc args =
          file :: files ->
             let inp, close_in = in_channel_of_any_value venv pos file in
             let inx = venv_find_channel venv pos inp in
-               line_loop inx;
+            let () =
+               try line_loop inx with
+                  exn when close_in ->
+                     venv_close_channel venv pos inp;
+                     raise exn
+            in
                if close_in then
                   venv_close_channel venv pos inp;
                file_loop files
        | [] ->
             ()
    in
-      file_loop files;
+   let return =
+      try file_loop files; ValNone with
+         Break (_, venv) ->
+            ValEnv (venv, ExportAll)
+   in
       Lm_channel.flush outx;
-      ValNone
+      return
 
 (*
  * \begin{doc}
- * \subsection{Lexer}
+ * \fun{lex}
+ *
+ * \begin{verbatim}
+ *    lex(files)
+ *    case pattern1
+ *       body1
+ *    case pattern2
+ *       body2
+ *    ...
+ *    default
+ *       bodyd
+ * \end{verbatim}
+ *
+ * The \verb+lex+ function provides a simple lexical-style scanner
+ * function.  The input is a sequence of files or channels.  The cases
+ * specify regular expressions.  Each time the input is read, the regular
+ * expression that matches the \emph{longest prefix} of the input is selected,
+ * and the body is evaluated.
+ *
+ * If two clauses both match the same input, the \emph{last} one is selected
+ * for execution.  The \verb+default+ case matches the regular expression \verb+.+;
+ * you probably want to place it first in the pattern list.
+ *
+ * If the body end with an \verb+export+ directive,
+ * the state is passed to the next clause.
+ *
+ * For example, the following program collects all occurrences of alphanumeric
+ * words in an input file.
+ *
+ * \begin{verbatim}
+ *     collect-words($(files)) =
+ *        words[] =
+ *        lex($(files))
+ *        default
+ *           # empty
+ *        case $"[[:alnum:]]+" g
+ *           words[] += $0
+ *           export
+ * \end{verbatim}
+ *
+ * The \verb+default+ case, if one exists, matches single characters.  Since
+ *
+ * It is an error if the input does not match any of the regular expressions.
+ *
+ * The \hyperfun{break} can be used to abort the loop.
+ * \end{doc}
+ *)
+let eof_sym = Lm_symbol.add "eof"
+
+let lex venv pos loc args =
+   let pos = string_pos "lex" pos in
+   let cases, files = awk_args venv pos loc args in
+
+   (* Add a clause for EOF *)
+   let _, lex = Lexer.add_clause Lexer.empty eof_sym "\\'" in
+
+   (* Get lexers for all the cases *)
+   let lex, cases, _ =
+      List.fold_left (fun (lex, cases, index) (v, test, body) ->
+            let args = values_of_value venv pos test in
+            let pattern =
+               match args with
+                  pattern :: _ ->
+                     string_of_value venv pos pattern
+                | [] ->
+                     ""
+            in
+            let pattern =
+               if Lm_symbol.eq v case_sym then
+                  pattern
+               else if Lm_symbol.eq v default_sym then
+                  "."
+               else
+                  raise (OmakeException (loc_pos loc pos, StringVarError ("unknown case", v)))
+            in
+               let action_sym = Lm_symbol.make "action" index in
+               let _, lex =
+                  try Lexer.add_clause lex action_sym pattern with
+                     Failure err ->
+                        let msg = sprintf "Malformed regular expression '%s'" pattern in
+                           raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+               in
+               let cases = SymbolTable.add cases action_sym body in
+                  lex, cases, succ index) (lex, SymbolTable.empty, 0) cases
+   in
+
+   (* Process the files *)
+   let rec input_loop venv inx =
+      let action_sym, lexeme_loc, lexeme, args = Lexer.lex lex inx in
+         if Lm_symbol.eq action_sym eof_sym then
+            venv
+         else
+            let venv = venv_add_match venv lexeme args in
+            let venv = venv_add_var venv ScopeProtected pos loc_sym (ValOther (ValLocation lexeme_loc)) in
+            let body =
+               try SymbolTable.find cases action_sym with
+                  Not_found ->
+                     raise (Invalid_argument "lex")
+            in
+            let result = eval_body_value venv pos body in
+            let venv, _ = add_exports venv pos result in
+               input_loop venv inx
+   in
+   let rec file_loop venv files =
+      match files with
+         file :: files ->
+            let inp, close_in = in_channel_of_any_value venv pos file in
+            let inx = venv_find_channel venv pos inp in
+            let venv =
+               try input_loop venv inx with
+                  (Break _ | Return _ | UncaughtException _ ) as exn ->
+                     if close_in then
+                        venv_close_channel venv pos inp;
+                     raise exn
+                | exn ->
+                     if close_in then
+                        venv_close_channel venv pos inp;
+                     raise (UncaughtException (pos, exn))
+            in
+               if close_in then
+                  venv_close_channel venv pos inp;
+               file_loop venv files
+       | [] ->
+            venv
+   in
+   let venv =
+      try file_loop venv files with
+         Break (_, venv) ->
+            venv
+   in
+      ValEnv (venv, ExportAll)
+
+(*
+ * \begin{doc}
+ * \fun{lex-search}
+ *
+ * \begin{verbatim}
+ *    lex-search(files)
+ *    case pattern1
+ *       body1
+ *    case pattern2
+ *       body2
+ *    ...
+ *    default
+ *       bodyd
+ * \end{verbatim}
+ *
+ * The \verb+lex-search+ function is like the \verb+lex+ function, but input that
+ * does not match any of the regular expressions is skipped.  If the clauses include
+ * a \verb+default+ case, then the \verb+default+ matches any skipped text.
+ *
+ * For example, the following program collects all occurrences of alphanumeric
+ * words in an input file, skipping any other text.
+ *
+ * \begin{verbatim}
+ *     collect-words($(files)) =
+ *        words[] =
+ *        lex-search($(files))
+ *        default
+ *           eprintln(Skipped $0)
+ *        case $"[[:alnum:]]+" g
+ *           words[] += $0
+ *           export
+ * \end{verbatim}
+ *
+ * The \verb+default+ case, if one exists, matches single characters.  Since
+ *
+ * It is an error if the input does not match any of the regular expressions.
+ *
+ * The \hyperfun{break} can be used to abort the loop.
+ * \end{doc}
+ *)
+let lex_search venv pos loc args =
+   let pos = string_pos "lex-search" pos in
+   let cases, files = awk_args venv pos loc args in
+
+   (* Get lexers for all the cases *)
+   let lex, cases, default, _ =
+      List.fold_left (fun (lex, cases, default, index) (v, test, body) ->
+            let args = values_of_value venv pos test in
+            let pattern =
+               match args with
+                  pattern :: _ ->
+                     string_of_value venv pos pattern
+                | [] ->
+                     ""
+            in
+               if Lm_symbol.eq v case_sym then
+                  let action_sym = Lm_symbol.make "action" index in
+                  let _, lex =
+                     try Lexer.add_clause lex action_sym pattern with
+                        Failure err ->
+                           let msg = sprintf "Malformed regular expression '%s'" pattern in
+                              raise (OmakeException (loc_pos loc pos, StringStringError (msg, err)))
+                  in
+                  let cases = SymbolTable.add cases action_sym body in
+                     lex, cases, default, succ index
+               else if Lm_symbol.eq v default_sym then
+                  lex, cases, Some body, index
+               else
+                  raise (OmakeException (loc_pos loc pos, StringVarError ("unknown case", v)))) (**)
+         (Lexer.empty, SymbolTable.empty, None, 0) cases
+   in
+
+   (* What to do for skipped text *)
+   let skip venv lexeme_loc lexeme =
+      match lexeme, default with
+         "", _
+       | _, None ->
+            venv
+       | _, Some body ->
+            let venv = venv_add_match venv lexeme [] in
+            let venv = venv_add_var venv ScopeProtected pos loc_sym (ValOther (ValLocation lexeme_loc)) in
+            let result = eval_body_value venv pos body in
+            let venv, _ = add_exports venv pos result in
+               venv
+   in
+
+   (* Process the files *)
+   let rec input_loop venv inx =
+      match Lexer.searchto lex inx with
+         Lexer.LexEOF ->
+            venv
+       | Lexer.LexSkipped (lexeme_loc, lexeme) ->
+            skip venv loc lexeme
+       | Lexer.LexMatched (action_sym, lexeme_loc, skipped, lexeme, args) ->
+            (* Process skipped text *)
+            let venv = skip venv lexeme_loc skipped in
+
+            (* Process the matched text *)
+            let venv = venv_add_match venv lexeme args in
+            let venv = venv_add_var venv ScopeProtected pos loc_sym (ValOther (ValLocation lexeme_loc)) in
+            let body =
+               try SymbolTable.find cases action_sym with
+                  Not_found ->
+                     raise (Invalid_argument "lex")
+            in
+            let result = eval_body_value venv pos body in
+            let venv, _ = add_exports venv pos result in
+               input_loop venv inx
+   in
+
+   (* Process each file *)
+   let rec file_loop venv files =
+      match files with
+         file :: files ->
+            let inp, close_in = in_channel_of_any_value venv pos file in
+            let inx = venv_find_channel venv pos inp in
+            let venv =
+               try input_loop venv inx with
+                  (Break _ | Return _ | UncaughtException _ ) as exn ->
+                     if close_in then
+                        venv_close_channel venv pos inp;
+                     raise exn
+                | exn ->
+                     if close_in then
+                        venv_close_channel venv pos inp;
+                     raise (UncaughtException (pos, exn))
+            in
+               if close_in then
+                  venv_close_channel venv pos inp;
+               file_loop venv files
+       | [] ->
+            venv
+   in
+   let venv =
+      try file_loop venv files with
+         Break (_, venv) ->
+            venv
+   in
+      ValEnv (venv, ExportAll)
+
+(*
+ * \begin{doc}
+ * \obj{Lexer}
  *
  * The \verb+Lexer+ object defines a facility for lexical analysis, similar to the
  * \Cmd{lex}{1} and \Cmd{flex}{1} programs.
@@ -925,21 +1539,8 @@ let lex_rule venv pos loc args =
             in
 
             (* Add the method *)
-            let venv =
-               match body with
-                  _ :: _ ->
-                     let body =
-                        SequenceExp (**)
-                           (loc, (LetVarExp (loc, ScopeProtected, name_sym, VarDefNormal, ConstString (loc, action_name))
-                                  :: LetVarExp (loc, ScopeProtected, val_sym, VarDefNormal, ConstString (loc, ""))
-                                  :: body))
-                     in
-                        venv_add_var venv ScopeProtected pos action_sym (ValFun (ArityExact 0, venv_get_env venv, [], body))
-                | [] ->
-                     venv
-            in
-
-            (* Add back the lexer *)
+            let body = SequenceExp (loc, body) in
+            let venv = venv_add_var venv ScopeProtected pos action_sym (ValFun (ArityExact 0, venv_get_env venv, [], body)) in
             let venv = venv_add_var venv ScopeProtected pos builtin_sym (ValOther (ValLexer lexer)) in
                ValEnv (venv, ExportAll)
 
@@ -1231,16 +1832,12 @@ let parse_build venv pos loc args =
  * Get the precedence option.
  *)
 let prec_option venv pos loc options =
-   match options with
-      [ValArray [optname; optval] as v] ->
-         if string_of_value venv pos optname = ":prec:" then
-            Some (Lm_symbol.add (string_of_value venv pos optval))
-         else
-            raise (OmakeException (pos, StringValueError ("illegal option", v)))
-    | [] ->
-         None
-    | v :: _ ->
-         raise (OmakeException (pos, StringValueError ("illegal option", v)))
+   venv_map_fold (fun pre optname optval ->
+         let s = string_of_value venv pos optname in
+            if s = ":prec:" then
+               Some (Lm_symbol.add (string_of_value venv pos optval))
+            else
+               raise (OmakeException (pos, StringValueError ("illegal option", optname)))) None options
 
 (*
  * Compute an action name that is not defined in the current object.
@@ -1257,7 +1854,7 @@ let parse_rule venv pos loc args =
    let pos = string_pos "parse-rule" pos in
    let action, head, rhs, options, body =
       match args with
-         [_; action; head; rhs; ValArray options; ValBody (_, SequenceExp (_, body))] ->
+         [_; action; head; rhs; ValMap options; ValBody (_, SequenceExp (_, body))] ->
             let action = string_of_value venv pos action in
             let head = string_of_value venv pos head in
                if head = "" then   (* Action name was omitted *)
@@ -1404,20 +2001,20 @@ let () =
        true, "parse-right",           parse_right,          ArityExact 1;
        true, "parse-nonassoc",        parse_nonassoc,       ArityExact 1;
        true, "parse-build",           parse_build,          ArityExact 1;
+       true, "scan",                  scan,                 ArityRange (1, 3);
        true, "awk",                   awk,                  ArityExact 3;
-       true, "fsubst",                fsubst,               ArityExact 3]
+       true, "fsubst",                fsubst,               ArityExact 3;
+       true, "lex",                   lex,                  ArityExact 3;
+       true, "lex-search",            lex_search,           ArityExact 3]
    in
    let builtin_info =
       { builtin_empty with builtin_funs = builtin_funs }
    in
       register_builtin builtin_info
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
