@@ -2,7 +2,9 @@
  * Builtin shell operations.
  *
  * \begin{doc}
- * \section{Shell functions}
+ * \input{omake-shell}
+ *
+ * \section{Basic builtin functions}
  * \end{doc}
  *
  * ----------------------------------------------------------------
@@ -12,8 +14,8 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -22,7 +24,11 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -96,12 +102,14 @@ let signal_of_string pos s =
  * Print some text.
  *
  * \begin{doc}
- * \subsection{echo}
+ * \fun{echo}
  *
  * The \verb+echo+ function prints a string.
  *
- * \verb+$(echo <args>)+
- * \verb+echo <args>+
+ * \begin{verbatim}
+ * $(echo <args>)
+ * echo <args>
+ * \end{verbatim}
  * \end{doc}
  *)
 let echo_fun venv pos loc args =
@@ -127,21 +135,7 @@ let echo_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{jobs}
- *
- * The \verb+jobs+ function prints a list of jobs.
- *
- * \verb+jobs+
- * \end{doc}
- *)
-let jobs_fun venv pos loc args =
-   let _pos = string_pos "jobs" pos in
-      Omake_shell_job.jobs venv;
-      ValNone
-
-(*
- * \begin{doc}
- * \subsection{cd}
+ * \fun{cd}
  *
  * The \verb+cd+ function changes the current directory.
  *
@@ -269,7 +263,22 @@ let cd_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{bg}
+ * \section{Job control builtin functions}
+ * \fun{jobs}
+ *
+ * The \verb+jobs+ function prints a list of jobs.
+ *
+ * \verb+jobs+
+ * \end{doc}
+ *)
+let jobs_fun venv pos loc args =
+   let _pos = string_pos "jobs" pos in
+      Omake_shell_job.jobs venv;
+      ValNone
+
+(*
+ * \begin{doc}
+ * \fun{bg}
  *
  * The \verb+bg+ function places a job in the background.
  *
@@ -293,7 +302,7 @@ let bg_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{fg}
+ * \fun{fg}
  *
  * The \verb+fg+ function brings a job to the foreground.
  *
@@ -307,7 +316,7 @@ let fg_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{stop}
+ * \fun{stop}
  *
  * The \verb+stop+ function suspends a job.
  *
@@ -321,7 +330,7 @@ let stop_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{wait}
+ * \fun{wait}
  *
  * The \verb+wait+ function waits for a job to finish.
  * If no process identifiers are given, the shell waits for
@@ -337,7 +346,7 @@ let wait_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{kill}
+ * \fun{kill}
  *
  * The \verb+kill+ function signals a job.
  *
@@ -367,7 +376,8 @@ let kill_fun venv pos loc args =
 
 (*
  * \begin{doc}
- * \subsection{history}
+ * \section{Command history}
+ * \fun{history}
  *
  * \begin{verbatim}
  *     $(history-index) : Int
@@ -414,7 +424,7 @@ let history venv pos loc args =
 
 let () =
    let builtin_vars =
-      ["history-file",   (fun _ -> ValNode (Node.intern no_mount_points PhonyProhibited Dir.root (Omake_state.history_file ())));
+      ["history-file",   (fun _ -> ValNode (Node.create_node no_mount_info Mount.empty Dir.root (Omake_state.history_file ())));
        "history-length", (fun _ -> ValInt 100);
        "CDPATH",         (fun _ -> ValArray [ValString "."])]
    in
