@@ -2,24 +2,32 @@
  * The Unix-style test command.
  * See the usage comment below.
  *
+ * \begin{doc}
+ * \section{File predicates}
+ * \end{doc}
+ *
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2005 Mojave Group, Caltech
+ * Copyright (C) 2005-2006 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -659,7 +667,7 @@ and parse_exp venv pos tokens =
  * Usage.
  *
  * \begin{doc}
- * \subsection{test}
+ * \fun{test}
  *
  * \begin{verbatim}
  *    test(exp) : Bool
@@ -727,6 +735,28 @@ and parse_exp venv pos tokens =
  * the \emph{string}.
  *
  * A \emph{file} is a \emph{string} that represents the name of a file.
+ *
+ * The syntax mirrors that of the \Cmd{test}{1} program.  If you are on a Unix system, the man page
+ * explains more.  Here are some examples.
+ *
+ * \begin{verbatim}
+ *     # Create an empty file
+ *     osh> touch foo
+ *     # Is the file empty?
+ *     osh> test(-e foo)
+ *     - : true
+ *     osh> test(! -e foo)
+ *     - : false
+ *     # Create another file
+ *     osh> touch boo
+ *     # Is the newer file newer?
+ *     osh> test(boo -nt foo)
+ *     - : true
+ *     # A more complex query
+ *     # boo is newer than foo, and foo is empty
+ *     osh> test(\( boo -nt foo \) -a -e foo)
+ *     - : true
+ * \end{verbatim}
  * \end{doc}
  *)
 let print_usage venv pos loc =
@@ -783,7 +813,7 @@ let test_cmd venv pos loc argv =
  * Usage.
  *
  * \begin{doc}
- * \subsection{find}
+ * \fun{find}
  *
  * \begin{verbatim}
  *    find(exp) : Node Array
@@ -806,8 +836,33 @@ let test_cmd venv pos loc argv =
  * additions.
  *
  * \begin{itemize}
- * \item    \verb+-name+ \emph{string} : The current file matches the regular expression.
+ * \item    \verb+-name+ \emph{string} : The current file matches the glob expression
+ * (see Section~\ref{section:globbing}).
  * \end{itemize}
+ *
+ * The \verb+find+ function performs a recursive scan of all subdirectories.
+ * The following call is being run from the root of the \verb+omake+ source directory.
+ *
+ * \begin{verbatim}
+ *     osh> find(. -name fo* )
+ *     - : <array
+ *             /home/jyh/.../omake/mk/.svn/format
+ *             /home/jyh/.../omake/RPM/.svn/format
+ *             ...
+ *             /home/jyh/.../omake/osx_resources/installer_files/.svn/format>
+ * \end{verbatim}
+ *
+ * Another example, listing only those files that are normal files
+ * or symbolic links.
+ *
+ * \begin{verbatim}
+ *     osh> find(. -name fo* -a \( -f {} -o -L {} \))
+ *     - : <array
+ *             /home/jyh/.../omake/mk/.svn/format
+ *             /home/jyh/.../omake/RPM/.svn/format
+ *             ...
+ *             /home/jyh/.../omake/osx_resources/installer_files/.svn/format>
+ * \end{verbatim}
  * \end{doc}
  *)
 let test venv pos loc args =
@@ -820,10 +875,7 @@ let test venv pos loc args =
        | _ ->
             raise (OmakeException (loc_pos loc pos, ArityMismatch (ArityExact 1, List.length args)))
    in
-      if code then
-         val_true
-      else
-         val_false
+      val_of_bool code
 
 let shell_test cmd venv pos loc args =
    let pos = string_pos "shell-test" pos in
@@ -941,12 +993,9 @@ let () =
    in
       register_builtin builtin_info
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
