@@ -8,17 +8,21 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -32,6 +36,10 @@
 #include <caml/custom.h>
 
 #ifdef WIN32
+#include <caml/signals.h>
+
+/* Disable some of the warnings */
+#pragma warning( disable : 4127 4189 4702 4706 4996 )
 
 #ifndef _WIN32_WINNT
 #define _WIN32_WINNT 0x0400
@@ -106,7 +114,7 @@
 #define MAX_PID                 29999
 
 /*
- * The process is grou leader if pid and pgrp are the same.
+ * The process is group leader if pid and pgrp are the same.
  */
 typedef struct _process {
     int pid;                            // Process identifier
@@ -118,7 +126,7 @@ typedef struct _process {
     unsigned status    : 2;             // Process status
     unsigned code      : 8;             // Exit code, if the process exited
     HANDLE handle;                      // The actual process (or event if this is a thread)
-    DWORD wid;				// Process identifier
+    DWORD wid;          // Process identifier
     struct _process *next;              // Linked list
 } Process;
 
@@ -133,7 +141,7 @@ typedef struct _shell_state {
 /*
  * The state.
  */
-static ShellState *state;
+static ShellState *state = NULL;
 
 /************************************************
  * Utilities.
@@ -176,11 +184,25 @@ static int string_escape_length(const char *strp)
     extra = 0;
     escaped = 0;
     while(c = *strp++) {
-        if(c <= ' ' || c == '\'')
-            escaped++;
-        else if(c == '"') {
+        switch(c) {
+        case '"':
             escaped++;
             extra++;
+            break;
+        case '*':
+        case '?':
+        case '[':
+        case ']':
+        case '{':
+        case '}':
+        case '~':
+        case '\'':
+            escaped++;
+            break;
+        default:
+            if(c <= ' ')
+                escaped++;
+            break;
         }
     }
     return escaped ? extra + 2 : 0;
@@ -191,7 +213,7 @@ static int string_escape_length(const char *strp)
  */
 static void string_copy_escaped(char *dstp, const char *srcp)
 {
-    int c;
+    char c;
 
     *dstp++ = '"';
     while(c = *srcp) {
@@ -372,10 +394,6 @@ static int resume_process(Process *processp)
  */
 static int kill_process(Process *processp)
 {
-    DWORD threads[MAX_THREAD_COUNT];
-    HANDLE handle;
-    int i, count;
-
     /*
      * Set the killed flag.
      * For threads, this will generate an exception
@@ -448,6 +466,7 @@ static int alloc_pid(void)
  * Fixup the process list, and
  * return a wait code.
  */
+#pragma warning( disable : 4100 )
 static value handle_wait(const char *debug, Process **processpp)
 {
     CAMLparam0();
@@ -494,8 +513,7 @@ static value handle_wait(const char *debug, Process **processpp)
 static void terminate_and_exit(void)
 {
     Process *processp;
-    int wait, i;
-    DWORD code;
+    int wait;
 
     /* Send all processes the termination signal */
     wait = 0;
@@ -560,7 +578,7 @@ value omake_shell_sys_suspend(value v_pgrp)
     fflush(stderr);
 #endif
     if(process_group_map(suspend_process, Int_val(v_pgrp)) < 0)
-        failwith("suspend");
+        failwith("omake_shell_sys_suspend");
     CAMLreturn(Val_unit);
 }
 
@@ -572,7 +590,7 @@ value omake_shell_sys_resume(value v_pgrp)
     fflush(stderr);
 #endif
     if(process_group_map(resume_process, Int_val(v_pgrp)) < 0)
-        failwith("suspend");
+        failwith("omake_shell_sys_resume");
     CAMLreturn(Val_unit);
 }
 
@@ -584,7 +602,7 @@ value omake_shell_sys_kill(value v_pgrp)
     fflush(stderr);
 #endif
     if(process_group_map(kill_process, Int_val(v_pgrp)) < 0)
-        failwith("suspend");
+        failwith("omake_shell_sys_kill");
     CAMLreturn(Val_unit);
 }
 
@@ -710,7 +728,7 @@ value omake_shell_sys_check_thread(value v_unit)
 {
     CAMLparam1(v_unit);
     Process *processp;
-    int id;
+    DWORD id;
 
 #ifdef OSH_DEBUG
     fprintf(stderr, "omake_shell_sys_check_thread\n");
@@ -745,8 +763,8 @@ static value omake_shell_sys_wait_aux(value v_pgrp, value v_leader, value v_noha
     int processes[MAXIMUM_WAIT_OBJECTS];
     HANDLE handles[MAXIMUM_WAIT_OBJECTS];
     Process **processpp, *processp;
-    int ncount, index, code, pid, pgrp, leader;
-    DWORD exitcode, timeout;
+    int ncount, code, pid, pgrp, leader;
+    DWORD exitcode, timeout, index;
 
 #ifdef OSH_DEBUG
     fprintf(stderr, "omake_shell_sys_wait\n");
@@ -790,7 +808,7 @@ static value omake_shell_sys_wait_aux(value v_pgrp, value v_leader, value v_noha
         /* Perform the wait */
         enter_blocking_section();
         index = WaitForMultipleObjects(ncount, handles, FALSE, timeout);
-        if(index == -1)
+        if(index == WAIT_FAILED)
             code = GetLastError();
         leave_blocking_section();
 
@@ -923,6 +941,10 @@ value omake_shell_sys_create_process(value v_info)
         index += length + white;
     }
     argv[index++] = 0;
+#if 0
+    fprintf(stderr, "Command: %s\n", argv);
+    fflush(stderr);
+#endif
 
     /* Get the directory */
     dir = String_val(Field(v_info, CREATE_PROCESS_DIR));
@@ -971,12 +993,34 @@ value omake_shell_sys_create_process(value v_info)
                            &startup,                       // Startup info
                            &process);                      // Process info
     if(status == FALSE) {
-#ifdef OSH_DEBUG
-        print_error("CreateProcess");
-#endif
-        fprintf(stderr, "Command: %s\n", command);
-        fflush(stderr);
-        failwith("omake_shell_sys_create_process: process creation failed");
+        char * lpMsgBuf = NULL;
+        int bufLen = FormatMessageA( 
+            FORMAT_MESSAGE_ALLOCATE_BUFFER | 
+            FORMAT_MESSAGE_FROM_SYSTEM | 
+            FORMAT_MESSAGE_IGNORE_INSERTS,
+            NULL,
+            GetLastError(),
+            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
+            (LPTSTR) &lpMsgBuf,
+            0,
+            NULL );
+
+        #ifdef OSH_DEBUG
+            print_error("CreateProcess");
+            fprintf(stderr, "Command: %s\n", command);
+            fflush(stderr);
+        #endif
+        if ((bufLen < 1) || (bufLen > 1024)) {
+            if (lpMsgBuf != NULL) 
+                LocalFree( lpMsgBuf );
+            failwith("omake_shell_sys_create_process: process creation failed");
+        } else {
+            char err[2048];
+            sprintf(err, "omake_shell_sys_create_process: process creation failed: %s", (char *)lpMsgBuf);
+            if (lpMsgBuf != NULL) 
+                LocalFree( lpMsgBuf );
+            failwith(err);
+        }
     }
     CloseHandle(process.hThread);
 
@@ -1130,7 +1174,7 @@ value omake_shell_sys_set_tty_pgrp(value v_pgrp)
     pgrp = Int_val(v_pgrp);
     if(pgrp == 0)
         pgrp = getpgrp();
-    if(tcsetpgrp (0, pgrp) < 0)
+    if(tcsetpgrp(0, pgrp) < 0)
         perror("tcsetpgrp");
     return Val_unit;
 }
@@ -1141,7 +1185,8 @@ value omake_shell_sys_setpgid(value v_pid, value v_pgrp)
 
     pid = Int_val(v_pid);
     pgrp = Int_val(v_pgrp);
-    setpgid(pid, pgrp);
+    if(setpgid(pid, pgrp) < 0)
+        perror("setpgid");
     return Val_unit;
 }
 
