@@ -4,12 +4,12 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2007 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -18,10 +18,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{nogin@metaprl.org}
  * @end[license]
  *)
 open Lm_printf
@@ -39,7 +43,8 @@ open Omake_lexer
 open Omake_parser
 open Omake_node_sig
 open Omake_exec_type
-open Omake_options_type
+open Omake_shell_type
+open Omake_options
 open Omake_command_type
 open Omake_ir_free_vars
 
@@ -54,6 +59,13 @@ val debug_implicit : bool ref
  *)
 type 'a source = node_kind * 'a
 
+(*
+ * A target value that represents a node in a rule.
+ *)
+type target =
+   TargetNode of Node.t
+ | TargetString of string
+
 (*
  * Type of objects.
  *)
@@ -107,26 +119,25 @@ type rule_kind =
    RuleNormal
  | RuleScanner
 
-(*
- * Execution service.
- *)
-type venv_command_inst = (exp, value) poly_command_inst
-and venv_command_line = (venv, exp, value) poly_command_line
-
 (*
  * Export kinds.
  *    ExportFile: this is the result of file evaluation
- *    ExportSymbols syms: export just the symbols in the list
+ *    ExportList: export just the listed items (variable values, implicit rules/dependencies, phony target set)
  *    ExportAll: export the entire environment.
+ *    ExportValue v: just like Exportall, but return the value too
  *)
 and export =
    ExportFile
  | ExportAll
  | ExportDir
- | ExportRules
- | ExportSymbols of symbol list
+ | ExportList of export_elt list
  | ExportValue of value
 
+and export_elt =
+   ExportSymbol of symbol
+ | ExportRules
+ | ExportPhonies
+
 (*
  * Possible values.
  * For the function, the obj is the static scope.
@@ -140,6 +151,7 @@ and value =
  | ValString      of string
  | ValData        of string
  | ValQuote       of value list
+ | ValQuoteString of char * value list
  | ValApply       of loc * scope_kind * var * value list
  | ValSuperApply  of loc * scope_kind * var * var * value list
  | ValMethodApply of loc * scope_kind * var list * value list
@@ -179,7 +191,7 @@ and value_other =
  *)
 and command =
    CommandSection of value * free_vars * exp    (* Name of the section, its free variables, and the expression *)
- | CommandValue of value
+ | CommandValue of loc * value
 
 and command_info =
    { command_env       : venv;
@@ -206,17 +218,54 @@ and erule =
 
 (*
  * A listing of all the explicit rules.
+ *
+ *    explicit_targets     : the collapsed rules for each explicit target
+ *    explicit_deps        : the table of explicit rules that are just dependencies
+ *    explicit_rules       : the table of all individual explicit rules
+ *    explicit_directories : the environment for each directory in the project
  *)
 and erule_info =
    { explicit_targets         : erule NodeTable.t;
      explicit_deps            : (NodeSet.t * NodeSet.t * NodeSet.t) NodeTable.t;   (* locks, sources, scanners *)
+     explicit_rules           : erule NodeMTable.t;
      explicit_directories     : venv DirTable.t
    }
 
 (*
- * Ordering info is abstract.
+ * Command lines.
  *)
-type ordering_info
+and arg_command_inst = (exp, arg_pipe, value) poly_command_inst
+and arg_command_line = (venv, exp, arg_pipe, value) poly_command_line
+
+and string_command_inst = (exp, string_pipe, value) poly_command_inst
+and string_command_line = (venv, exp, string_pipe, value) poly_command_line
+
+and apply        = venv -> Unix.file_descr -> Unix.file_descr -> Unix.file_descr -> (symbol * string) list -> value list -> int * value
+
+and value_cmd    = (unit, value list, value list) poly_cmd
+and value_apply  = (value list, value list, apply) poly_apply
+and value_group  = (unit, value list, value list, value list, apply) poly_group
+and value_pipe   = (unit, value list, value list, value list, apply) poly_pipe
+
+and arg_cmd      = (arg cmd_exe, arg, arg) poly_cmd
+and arg_apply    = (value, arg, apply) poly_apply
+and arg_group    = (arg cmd_exe, arg, value, arg, apply) poly_group
+and arg_pipe     = (arg cmd_exe, arg, value, arg, apply) poly_pipe
+
+and string_cmd   = (simple_exe, string, string) poly_cmd
+and string_apply = (value, string, apply) poly_apply
+and string_group = (simple_exe, string, value, string, apply) poly_group
+and string_pipe  = (simple_exe, string, value, string, apply) poly_pipe
+
+(*
+ * Command line parsing.
+ *)
+type lexer = string -> int -> int -> int option
+
+type tok =
+   TokString of value
+ | TokToken  of string
+ | TokGroup  of tok list
 
 (*
  * Type of execution servers.
@@ -226,7 +275,12 @@ type pid =
  | ExternalPid of int
  | ResultPid of int * value
 
-type exec = (venv_command_line, pid, value) Exec.t
+type exec = (arg_command_line, pid, value) Exec.t
+
+(*
+ * Ordering info is abstract.
+ *)
+type ordering_info
 
 (*
  * Inclusion scope is usually Pervasives,
@@ -246,6 +300,11 @@ type ir = symbol list * scope_kind SymbolTable.t * Omake_ir.exp
  *)
 val commands_are_trivial : command_info list -> bool
 
+(*
+ * Convert a target to a raw string.
+ *)
+val string_of_target : venv -> target -> string
+
 (*
  * This takes the starting directory.
  *)
@@ -287,28 +346,31 @@ val venv_add_file         : venv -> Node.t -> venv
 val venv_intern           : venv -> phony_ok -> string -> Node.t
 val venv_intern_cd        : venv -> phony_ok -> Dir.t -> string -> Node.t
 val venv_intern_dir       : venv -> string -> Dir.t
+val venv_intern_target    : venv -> phony_ok -> target -> Node.t
+val venv_dirname          : venv -> Dir.t -> string
+val venv_nodename         : venv -> Node.t -> string
 
 val venv_mount       : venv -> mount_option list -> Dir.t -> Dir.t -> venv
 
 val venv_add_var     : venv -> scope_kind -> pos -> var -> value -> venv
 val venv_add_args    : venv -> pos -> loc -> env -> var list -> value list -> venv
 val venv_add_assoc   : venv -> (var * string) list -> venv
-val venv_add_phony   : venv -> loc -> string list -> venv
+val venv_add_phony   : venv -> loc -> target list -> venv
 
 val venv_add_args_hack : venv -> pos -> loc -> env -> var list -> value list -> venv
 
-val venv_add_wild_match  : venv -> string -> venv
+val venv_add_wild_match  : venv -> value -> venv
 val venv_add_match_args  : venv -> string list -> venv
 val venv_add_match       : venv -> string -> string list -> venv
 val venv_explicit_target : venv -> Node.t -> venv
 
 val venv_add_rule : venv -> pos -> loc ->
    rule_multiple ->                     (* multiple, scanner, etc *)
-   string list ->                       (* targets *)
-   string list ->                       (* patterns *)
-   string source list ->                (* effects *)
-   string source list ->                (* sources *)
-   string source list ->                (* scanners *)
+   target list ->                       (* targets *)
+   target list ->                       (* patterns *)
+   target source list ->                (* effects *)
+   target source list ->                (* sources *)
+   target source list ->                (* scanners *)
    value list ->                        (* additional values the target depends on *)
    command list ->                      (* commands *)
    venv * erule list
@@ -319,6 +381,7 @@ val venv_add_rule : venv -> pos -> loc ->
 val venv_environment : venv -> string SymbolTable.t
 val venv_setenv : venv -> var -> string -> venv
 val venv_getenv : venv -> var -> string
+val venv_unsetenv : venv -> var -> venv
 val venv_defined_env : venv -> var -> bool
 
 (*
@@ -335,6 +398,7 @@ val venv_defined      : venv -> scope_kind -> var -> bool
 val venv_get_var      : venv -> scope_kind -> pos -> var -> value
 val venv_find_var     : venv -> scope_kind -> pos -> loc -> var -> value
 val venv_find_var_exn : venv -> scope_kind -> var -> value
+val venv_find_object_or_empty : venv -> scope_kind -> symbol -> obj
 
 (*
  * Static environments.
@@ -369,7 +433,7 @@ val venv_stdin            : prim_channel
 val venv_stdout           : prim_channel
 val venv_stderr           : prim_channel
 
-val venv_add_channel      : venv -> string -> Lm_channel.kind -> Lm_channel.mode -> bool -> Unix.file_descr -> prim_channel
+val venv_add_channel      : venv -> Lm_channel.t -> prim_channel
 val venv_close_channel    : venv -> pos -> prim_channel -> unit
 val venv_find_channel     : venv -> pos -> prim_channel -> Lm_channel.t
 val venv_find_channel_id  : venv -> pos -> int -> prim_channel
@@ -410,6 +474,7 @@ val venv_add_object_file      : venv -> Node.t -> obj -> unit
  *)
 val venv_map_empty       : map
 val venv_map_add         : map -> pos -> value -> value -> map
+val venv_map_remove      : map -> pos -> value -> map
 val venv_map_find        : map -> pos -> value -> value
 val venv_map_mem         : map -> pos -> value -> bool
 val venv_map_iter        : (value -> value -> unit) -> map -> unit
@@ -438,9 +503,9 @@ val venv_find_implicit_rules : venv -> Node.t -> erule list
 (*
  * Ordering.
  *)
-val venv_add_orders        : venv -> string list -> venv
+val venv_add_orders        : venv -> loc -> target list -> venv
 val venv_is_order          : venv -> string -> bool
-val venv_add_ordering_rule : venv -> pos -> loc -> var -> string -> string list -> venv
+val venv_add_ordering_rule : venv -> pos -> loc -> var -> target -> target list -> venv
 val venv_get_ordering_info : venv -> var -> ordering_info
 val venv_get_ordering_deps : venv -> ordering_info -> NodeSet.t -> NodeSet.t
 
@@ -470,6 +535,21 @@ val venv_find_target_is_buildable_proper_exn : venv -> Node.t -> bool
 val venv_add_target_is_buildable : venv -> Node.t -> bool -> unit
 val venv_add_target_is_buildable_proper : venv -> Node.t -> bool -> unit
 
+(*
+ * Printing.
+ *)
+val pp_print_tok : formatter -> tok -> unit
+
+val pp_print_string_pipe : formatter -> string_pipe -> unit
+val pp_print_string_command_inst : formatter -> string_command_inst -> unit
+val pp_print_string_command_line : formatter -> string_command_line -> unit
+val pp_print_string_command_lines : formatter -> string_command_line list -> unit
+
+val pp_print_arg_pipe : formatter -> arg_pipe -> unit
+val pp_print_arg_command_inst : formatter -> arg_command_inst -> unit
+val pp_print_arg_command_line : formatter -> arg_command_line -> unit
+val pp_print_arg_command_lines : formatter -> arg_command_line list -> unit
+
 (************************************************************************
  * For squashing (producing digests).
  *)
@@ -493,6 +573,7 @@ type omake_error =
  | StringIntError    of string * int
  | StringMethodError of string * var list
  | StringValueError  of string * value
+ | StringTargetError of string * target
  | LazyError         of (formatter -> unit)
  | UnboundVar        of var
  | UnboundFun        of var
@@ -501,6 +582,7 @@ type omake_error =
  | NotImplemented    of string
  | UnboundKey        of string
  | UnboundValue      of value
+ | NullCommand
 
 (*
  * General exception includes debugging info.
@@ -510,11 +592,13 @@ exception UncaughtException of pos * exn
 exception RaiseException    of pos * obj
 exception ExitException     of pos * int
 exception Return            of loc * value
+exception Break             of loc * venv
 
 (*
  * Omake's internal version of the Invalid_argument
  *)
 exception OmakeFatal of string
+exception OmakeFatalErr of pos * omake_error
 
 (*
  * Module for creating positions.
@@ -572,9 +656,10 @@ val debug_db : bool ref
 (*
  * Static loading.
  *)
-module Static :
+module type StaticSig =
 sig
-   type t
+   type in_handle
+   type out_handle
 
    (*
     * Open a file.  The Node.t is the name of the _source_ file,
@@ -582,8 +667,12 @@ sig
     * goes on our own.  Raises Not_found if the source file
     * can't be found.
     *)
-   val create     : venv -> Node.t -> t
-   val close      : t -> unit
+   val create_in    : venv -> Node.t -> in_handle
+   val close_in     : in_handle -> unit
+
+   val create_out   : venv -> Node.t -> out_handle
+   val recreate_out : in_handle -> out_handle
+   val close_out    : out_handle -> unit
 
    (*
     * Unfortunately, the IR type is delayed because it
@@ -591,27 +680,30 @@ sig
     * Omake_ir_ast depends on this file.
     *)
 
-   (*
-    * Add the three kinds of entries.
-    *)
-   val add_ir     : t -> ir -> unit
-   val add_object : t -> obj -> unit
-   val add_values : t -> obj -> unit
-
    (*
     * Fetch the three kinds of entries.
     *)
-   val find_ir     : t -> ir
-   val find_object : t -> obj
-   val find_values : t -> obj
+   val find_ir     : in_handle -> ir
+   val find_object : in_handle -> obj
+   val find_values : in_handle -> obj SymbolTable.t
+
+   (*
+    * Add the three kinds of entries.
+    *)
+   val get_ir      : out_handle -> ir
+   val get_object  : out_handle -> obj
+   val get_values  : out_handle -> obj SymbolTable.t
+
+   val add_ir      : out_handle -> ir -> unit
+   val add_object  : out_handle -> obj -> unit
+   val add_values  : out_handle -> obj SymbolTable.t -> unit
 end;;
 
-(*!
- * @docoff
- *
+module Static : StaticSig;;
+
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
