@@ -8,17 +8,21 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -107,6 +111,12 @@ let arg_next =
 let arg_start =
    subtract arg_next white
 
+let neq_arg_next =
+   subtract arg_next ["TokEq"]
+
+let neq_arg_start =
+   subtract arg_start ["TokEq"]
+
 let other_start =
    subtract tokens ("TokWhite" :: "quote" :: List.flatten [id; colon])
 
@@ -137,6 +147,8 @@ let productions =
      "keyword_target_start",            keyword_target_start;
      "arg_next",                        arg_next;
      "arg_start",                       arg_start;
+     "neq_arg_next",                    neq_arg_next;
+     "neq_arg_start",                   neq_arg_start;
      "other_start",                     other_start;
      "other_method_id_white",           other_method_id_white;
      "other_method_id",                 other_method_id;
