@@ -8,8 +8,8 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -18,7 +18,11 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -62,11 +66,16 @@ type senv_globals =
      mutable senv_undefined : SymbolSet.t
    }
 
+(*
+ * XXX: HACK: JYH: the indexes need to be unique, so use a reference cell
+ * for the index.  This is done properly in 0.9.8.3+, so this is only
+ * a temporary hack.
+ *)
 type senv =
    { senv_class_names    : symbol list;
      senv_mode           : bool * scope_kind;
      senv_vars           : senv_var_scope;
-     senv_index          : int;
+     senv_index          : int ref;
      senv_id             : symbol;
      senv_globals        : senv_globals
    }
@@ -136,7 +145,7 @@ let senv_create open_file vars node =
       { senv_class_names = [];
         senv_mode        = false, ScopeGlobal;
         senv_vars        = vars;
-        senv_index       = 0;
+        senv_index       = ref 0;
         senv_id          = Lm_symbol.make static_string 0;
         senv_globals     = globals
       }
@@ -246,6 +255,7 @@ let build_literal_argv e pos =
        | Omake_ast.StringExp (s, _) ->
             Buffer.add_string buf s
        | Omake_ast.QuoteExp (el, _)
+       | Omake_ast.QuoteStringExp (_, el, _)
        | Omake_ast.SequenceExp (el, _) ->
             collect_exp_list el
        | Omake_ast.ApplyExp (_, _, _, loc)
@@ -289,6 +299,8 @@ let rec build_string senv e pos =
             ConstString (loc, s)
        | Omake_ast.QuoteExp (el, loc) ->
             build_quote_string senv el pos loc
+       | Omake_ast.QuoteStringExp (c, el, loc) ->
+            build_quote_string_string senv c el pos loc
        | Omake_ast.SequenceExp ([e], _) ->
             build_string senv e pos
        | Omake_ast.SequenceExp (el, loc) ->
@@ -341,6 +353,11 @@ and build_quote_string senv el pos loc =
    let args = build_sequence_string_aux senv el pos loc in
       QuoteString (loc, args)
 
+and build_quote_string_string senv c el pos loc =
+   let pos = string_pos "build_quote_string_string" pos in
+   let args = build_sequence_string_aux senv el pos loc in
+      QuoteStringString (loc, c, args)
+
 and build_sequence_string_aux senv el pos loc =
    let pos = string_pos "build_sequence_string_aux" pos in
    let buf = Buffer.create 32 in
@@ -390,6 +407,7 @@ and build_sequence_string_aux senv el pos loc =
                 | ArrayString _
                 | ArrayOfString _
                 | QuoteString _
+                | QuoteStringString _
                 | ExpString _
                 | CasesString _
                 | ThisString _ ->
@@ -493,7 +511,8 @@ and build_exp senv e =
       match e with
          Omake_ast.NullExp loc
        | Omake_ast.StringExp (_, loc)
-       | Omake_ast.QuoteExp (_, loc) ->
+       | Omake_ast.QuoteExp (_, loc)
+       | Omake_ast.QuoteStringExp (_, _, loc) ->
             senv, SequenceExp (loc, [])
        | Omake_ast.SequenceExp ([e], _)
        | Omake_ast.BodyExp ([e], _) ->
@@ -526,7 +545,20 @@ and build_exp senv e =
        | Omake_ast.FunDefExp (v, params, e, loc) ->
             build_fun_def_exp senv v params e pos loc
        | Omake_ast.RuleExp (multiple, target, pattern, source, commands, loc) ->
-            build_rule_exp senv multiple target pattern source commands pos loc
+            (* XXX: HACK: for compatibility with 0.9.8.3+, recognize .STATIC here *)
+            begin match target with
+               Omake_ast.SequenceExp ([Omake_ast.StringExp (".", _); Omake_ast.StringExp ("STATIC", _)], _) ->
+                  if multiple then
+                     raise (OmakeException (pos, StringError ".STATIC: rules can not use the :: syntax"));
+                  begin match pattern with
+                     Omake_ast.NullExp _ when SymbolTable.is_empty source ->
+                        build_static_scope_exp senv commands pos loc
+                   | _ ->
+                        raise (OmakeException (pos, StringError ".STATIC: rule dependencies not yet supported in this OMake version"))
+                  end
+             | _ ->
+                  build_rule_exp senv multiple target pattern source commands pos loc
+            end
        | Omake_ast.ShellExp (e, loc) ->
             build_shell_exp senv e pos loc
        | Omake_ast.CatchExp (_, _, _, loc) ->
@@ -558,6 +590,14 @@ and build_sequence senv pos rval el =
          let senv, el = build_sequence senv pos rval el in
             senv, e :: el
 
+    | Omake_ast.CommandExp (v, e, body, loc)
+      :: el when Lm_symbol.eq v while_sym ->
+         let cases, el = collect_cases [] el in
+         let pos = loc_pos loc pos in
+         let senv, e = build_opt_cases_command_exp senv v e cases body pos loc in
+         let senv, el = build_sequence senv pos rval el in
+            senv, e :: el
+
     | Omake_ast.CommandExp (v, arg, body, loc) :: el ->
          let cases, el = collect_cases [] el in
          let pos = loc_pos loc pos in
@@ -597,11 +637,11 @@ and build_scope_exp senv scope el pos loc =
 
 and build_static_scope_exp senv el pos loc =
    let pos = string_pos "build_scope_exp" pos in
-   let { senv_globals = globals; senv_index = index } = senv in
-   let { senv_file = file } = globals in
-   let index = succ index in
+   let file = senv.senv_globals.senv_file in
+   let index = !(senv.senv_index) + 1 in
+   let () = senv.senv_index := index in
    let id = Lm_symbol.make static_string index in
-   let senv = { senv with senv_index = index; senv_id = id } in
+   let senv = { senv with senv_id = id } in
    let senv, scope = senv_scope senv (true, ScopeProtected) in
    let senv, el = build_sequence senv pos (fun senv -> senv, [ReturnObjectExp (loc, [])]) el in
    let senv, _ = senv_scope senv scope in
@@ -691,6 +731,17 @@ and build_cases_command_exp senv v arg cases commands pos loc =
             in
                build_cases_apply_exp senv v [arg] cases pos loc
 
+and build_opt_cases_command_exp senv v arg cases commands pos loc =
+   let pos = string_pos "build_opt_cases_command_exp" pos in
+   let cases =
+      if commands = [] then
+         cases
+      else
+         let default = default_sym, Omake_ast.NullExp loc, commands in
+            cases @ [default]
+   in
+      build_cases_apply_exp senv v [arg] cases pos loc
+
 (*
  * Commands.
  *)
@@ -714,6 +765,8 @@ and build_command_exp senv v arg commands pos loc =
          build_open_exp senv arg pos loc
       else if Lm_symbol.eq v declare_sym then
          build_declare_exp senv arg pos loc
+      else if Lm_symbol.eq v autoload_sym then
+         senv, SequenceExp (loc, [])
       else
          build_apply_exp senv v [arg] pos loc
 
@@ -959,12 +1012,16 @@ and build_rule_exp senv multiple target pattern sources body pos loc =
             if Lm_symbol.eq v normal_sym then
                options
             else
-               let option =
-                  ArrayString (loc, [ConstString (loc, Lm_symbol.to_string v); build_string senv source pos])
-               in
-                  option :: options) [] sources
+               let key = ConstString (loc, Lm_symbol.to_string v) in
+               let value = build_string senv source pos in
+                  key :: value :: options) [] sources
    in
-   let options = ArrayString (loc, options) in
+   let create_map_sym =
+      match options with
+         [] -> empty_map_sym
+       | _ -> create_lazy_map_sym
+   in
+   let options = ApplyString (loc, EagerApply, ScopeProtected, create_map_sym, options) in
 
    (* Add the sources *)
    let source =
