@@ -4,12 +4,12 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2007 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -18,10 +18,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{nogin@metaprl.org}
  * @end[license]
  *)
 open Lm_debug
@@ -45,8 +49,6 @@ open Omake_exec_type
 open Omake_exec_util
 open Omake_cache_type
 open Omake_command_type
-open Omake_command_digest
-open Omake_options_type
 open Omake_symbol
 
 module Pos = MakePos (struct let name = "Omake_eval" end)
@@ -98,6 +100,13 @@ type include_flag =
  * Utilities.
  *)
 
+(*
+ * Add an optional quote.
+ *)
+let buffer_add_quote buf = function
+   Some c -> Buffer.add_char buf c
+ | None -> ()
+
 (*
  * The various forms of empty values.
  *)
@@ -120,6 +129,7 @@ let rec is_empty_value v =
     | ValFloat _
     | ValData _
     | ValQuote _
+    | ValQuoteString _
     | ValString _
     | ValArray _
     | ValApply _
@@ -167,6 +177,7 @@ let rec is_array_value v =
     | ValFloat _
     | ValData _
     | ValQuote _
+    | ValQuoteString _
     | ValString _
     | ValApply _
     | ValImplicit _
@@ -204,10 +215,22 @@ let strategy_is_eager be_eager strategy scope =
          else
             scope = ScopePrivate
 
+let eval_export_elt pos = function
+   ".RULE" ->
+      ExportRules
+ | ".PHONY" ->
+      ExportPhonies
+ | "" ->
+      raise (OmakeException(pos, StringError ("attempted to export an empty-named variable")))
+ | s ->
+      if s.[0] = '.' then
+         raise (OmakeException(pos, StringStringError ("unknown export special", s)))
+      else
+         ExportSymbol (Lm_symbol.add s)
+
 (************************************************************************
  * Compiling utilities.
  *)
-
 let postprocess_ir ir =
    let () =
       if debug print_ir then
@@ -328,11 +351,11 @@ and find_include_file venv pos loc filename =
                         raise (OmakeException (loc_pos loc pos, LazyError print_error))
       else
          let dirname = Filename.dirname filename in
-         let filename = Filename.basename filename in
-         let fullname = filename ^ omake_file_suffix in
+         let basename = Filename.basename filename in
+         let fullname = basename ^ omake_file_suffix in
          let path = venv_find_var venv ScopeGlobal pos loc omakepath_sym in
-         let path = values_of_value venv pos path in
-         let path = path_of_values venv pos path dirname in
+         let full_path = values_of_value venv pos path in
+         let path = path_of_values venv pos full_path dirname in
          let cache = venv_cache venv in
          let listing = Omake_cache.ls_path cache path in
             try
@@ -344,14 +367,19 @@ and find_include_file venv pos loc filename =
             with
                Not_found ->
                   try
-                     match Omake_cache.listing_find cache listing filename with
+                     match Omake_cache.listing_find cache listing basename with
                         DirEntry dir ->
                            raise (OmakeException (loc_pos loc pos, StringDirError ("is a directory", dir)))
                       | NodeEntry node ->
                            node
                   with
                      Not_found ->
-                        raise (OmakeException (loc_pos loc pos, StringStringError ("include file not found", filename)))
+                        let print_error buf =
+                           fprintf buf "@[<hv 3>include file %s not found in OMAKEPATH@ (@[<hv3>OMAKEPATH[] =%a@])@]" (**)
+                              filename
+                              pp_print_value_list full_path
+                        in
+                           raise (OmakeException (loc_pos loc pos, LazyError print_error))
 
 and open_ir venv filename pos loc =
    let pos = string_pos "open_ir" pos in
@@ -370,14 +398,21 @@ and open_ir venv filename pos loc =
  * Try to load the old entry.
  * If it fails, compile the file and save the new entry.
  *)
-and compile_ir_info venv scope info pos loc source =
+and compile_add_ir_info venv scope info pos loc source =
    let _pos = string_pos "compile_ir_info" pos in
-      try Static.find_ir info with
+      try Static.get_ir info with
          Not_found ->
             let ir = parse_ir venv scope source in
                Static.add_ir info ir;
                ir
 
+and compile_ir_info venv scope info pos loc source =
+   let _pos = string_pos "compile_ir_info" pos in
+      try Static.find_ir info with
+         Not_found ->
+            let info = Static.recreate_out info in
+               compile_add_ir_info venv scope info pos loc source
+
 and compile_ir venv scope pos loc source =
    let pos = string_pos "compile_ir" pos in
       (*
@@ -390,12 +425,12 @@ and compile_ir venv scope pos loc source =
                 * Open the database.
                 *)
                let info =
-                  try Static.create venv source with
+                  try Static.create_in venv source with
                      Not_found ->
                         raise (OmakeException (loc_pos loc pos, StringNodeError ("can't open IR", source)))
                in
                let ir = compile_ir_info venv scope info pos loc source in
-                  Static.close info;
+                  Static.close_in info;
                   ir
             in
                venv_add_ir_file venv source ir;
@@ -404,6 +439,21 @@ and compile_ir venv scope pos loc source =
 (*
  * The object file contains the evaluated file.
  *)
+and compile_add_object_info compile venv pos info source =
+   let _pos = string_pos "compile_add_object_info_info" pos in
+   let info = Static.recreate_out info in
+      try Static.get_object info with
+         Not_found ->
+            let obj = compile info source in
+               Static.add_object info obj;
+               obj
+
+and compile_object_info compile venv pos info source =
+   let _pos = string_pos "compile_object_info" pos in
+      try Static.find_object info with
+         Not_found ->
+            compile_add_object_info compile venv pos info source
+
 and compile_object compile venv pos loc source =
    let pos = string_pos "compile_ast" pos in
 
@@ -417,7 +467,7 @@ and compile_object compile venv pos loc source =
                 * Open the database.
                 *)
                let info =
-                  try Static.create venv source with
+                  try Static.create_in venv source with
                      Not_found ->
                         raise (OmakeException (loc_pos loc pos, StringNodeError ("can't open object", source)))
                in
@@ -426,14 +476,8 @@ and compile_object compile venv pos loc source =
                 * Try to load the old entry.
                 * If it fails, compile the file and save the new entry.
                 *)
-               let obj =
-                  try Static.find_object info with
-                     Not_found ->
-                        let obj = compile info source in
-                           Static.add_object info obj;
-                           obj
-               in
-                  Static.close info;
+               let obj = compile_object_info compile venv pos info source in
+                  Static.close_in info;
                   obj
             in
                venv_add_object_file venv source obj;
@@ -445,13 +489,11 @@ and compile_object compile venv pos loc source =
 
 (*
  * Get the string representation of a value.
- * It is now not legal to convert an array to a string.
- * Hopefully we will catch all the internal uses appropriately.
+ * It not legal to convert an array to a string.
  *)
 and string_of_value venv pos v =
    let pos = string_pos "string_of_value" pos in
    let scratch_buf = Buffer.create 32 in
-   let dir1 = venv_dir venv in
    let rec collect v =
       match eval_prim_value venv pos v with
          (* Values that expand to nothing *)
@@ -474,6 +516,10 @@ and string_of_value venv pos v =
             List.iter collect vl
        | ValQuote vl ->
             string_of_quote_buf scratch_buf venv pos vl
+       | ValQuoteString (c, vl) ->
+            Buffer.add_char scratch_buf c;
+            string_of_quote_buf scratch_buf venv pos vl;
+            Buffer.add_char scratch_buf c
        | ValArray [v] ->
             collect v
        | ValArray vl ->
@@ -495,9 +541,9 @@ and string_of_value venv pos v =
        | ValString s ->
             Buffer.add_string scratch_buf s
        | ValDir dir2 ->
-            Buffer.add_string scratch_buf (Dir.name dir1 dir2)
+            Buffer.add_string scratch_buf (venv_dirname venv dir2)
        | ValNode node ->
-            Buffer.add_string scratch_buf (Node.name dir1 node)
+            Buffer.add_string scratch_buf (venv_nodename venv node)
        | ValKey _
        | ValApply _
        | ValImplicit _
@@ -512,15 +558,16 @@ and string_of_value venv pos v =
  * Collect the values in a quotation into a string.
  * Even array values are flattened without warning.
  *)
-and string_of_quote venv pos vl =
+and string_of_quote venv pos c vl =
    let pos = string_pos "string_of_quote" pos in
    let scratch_buf = Buffer.create 32 in
+      buffer_add_quote scratch_buf c;
       string_of_quote_buf scratch_buf venv pos vl;
+      buffer_add_quote scratch_buf c;
       Buffer.contents scratch_buf
 
 and string_of_quote_buf scratch_buf venv pos vl =
    let pos = string_pos "string_of_quote_buf" pos in
-   let dir1 = venv_dir venv in
    let rec collect v =
       match eval_value venv pos v with
          (* Values that expand to nothing *)
@@ -542,6 +589,10 @@ and string_of_quote_buf scratch_buf venv pos vl =
        | ValSequence vl
        | ValQuote vl ->
             List.iter collect vl
+       | ValQuoteString (c, vl) ->
+            Buffer.add_char scratch_buf c;
+            List.iter collect vl;
+            Buffer.add_char scratch_buf c
        | ValArray [v] ->
             collect v
        | ValArray vl ->
@@ -554,9 +605,9 @@ and string_of_quote_buf scratch_buf venv pos vl =
        | ValString s ->
             Buffer.add_string scratch_buf s
        | ValDir dir2 ->
-            Buffer.add_string scratch_buf (Dir.name dir1 dir2)
+            Buffer.add_string scratch_buf (venv_dirname venv dir2)
        | ValNode node ->
-            Buffer.add_string scratch_buf (Node.name dir1 node)
+            Buffer.add_string scratch_buf (venv_nodename venv node)
        | ValKey _
        | ValApply _
        | ValSuperApply _
@@ -592,8 +643,12 @@ and values_of_value venv pos v =
    (*
     * Convert a catenable value to a string
     *)
-   let bundle vl = ValQuote vl in
-   let tokens = Lm_string_util.tokens_create (fun s -> ValData s) bundle in
+   let group tokens  = ValSequence tokens in
+   let wrap_string s = ValString s in
+   let wrap_data s   = ValData s in
+   let wrap_token s  = ValData s in
+   let lexer _ _ _   = None in
+   let tokens = Lm_string_util.tokens_create_lexer ~lexer ~wrap_string ~wrap_data ~wrap_token ~group in
 
    (*
     * Array elements are always separate values.
@@ -640,7 +695,8 @@ and values_of_value venv pos v =
                  | ValFloat _
                  | ValDir _
                  | ValNode _
-                 | ValQuote _ ->
+                 | ValQuote _
+                 | ValQuoteString _ ->
                       collect (Lm_string_util.tokens_add tokens v) vl vll
 
                    (* Atomic values *)
@@ -680,6 +736,177 @@ and strings_of_value venv pos v =
    let values = values_of_value venv pos v in
       List.map (string_of_value venv pos) values
 
+(*
+ * Get a list of tokens from the value.
+ * This is a lot like the previous function, but we use a lexer
+ * for parsing special character sequences.
+ *)
+and tokens_of_value venv pos lexer v =
+   let pos = string_pos "tokens_of_value" pos in
+
+   (*
+    * Convert a catenable value to a string
+    *)
+   let group tokens  = TokGroup tokens in
+   let wrap_string s = TokString (ValString s) in
+   let wrap_data s   = TokString (ValData s) in
+   let wrap_token s  = TokToken s in
+   let tokens = Lm_string_util.tokens_create_lexer ~lexer ~wrap_string ~wrap_data ~wrap_token ~group in
+
+   (*
+    * Array elements are always separate values.
+    * The arrays are flattened.
+    *)
+   let rec collect_array (tokens : tok Lm_string_util.tokens) vl vll =
+      match vl, vll with
+         v :: vl, _ ->
+            (match eval_value venv pos v with
+                ValArray el ->
+                   collect_array tokens el (vl :: vll)
+              | ValSequence [v] ->
+                   collect_array tokens (v :: vl) vll
+              | v ->
+                   collect_array (Lm_string_util.tokens_atomic tokens (TokString v)) vl vll)
+       | [], vl :: vll ->
+            collect_array tokens vl vll
+       | [], [] ->
+            tokens
+   in
+
+   (*
+    * Collect_string is used when we have seen whitespace
+    * in a sequence.  Collect the values into the string buffer,
+    * then parse the string into separate tokens.
+    *)
+   let rec collect (tokens : tok Lm_string_util.tokens) vl vll =
+      match vl, vll with
+         v :: vl, _ ->
+            let v = eval_catenable_value venv pos v in
+               (match v with
+                   ValNone ->
+                      collect tokens vl vll
+
+                   (* Strings *)
+                 | ValString s ->
+                      collect (Lm_string_util.tokens_lex tokens s) vl vll
+                 | ValSequence el ->
+                      collect tokens el (vl :: vll)
+
+                   (* Other catenable values *)
+                 | ValData _
+                 | ValInt _
+                 | ValFloat _
+                 | ValDir _
+                 | ValNode _
+                 | ValQuote _ ->
+                      collect (Lm_string_util.tokens_add tokens (TokString v)) vl vll
+                 | ValQuoteString (_, v) ->
+                      collect (Lm_string_util.tokens_add tokens (TokString (ValQuote v))) vl vll
+
+                   (* Atomic values *)
+                 | ValArray el ->
+                      collect (collect_array (Lm_string_util.tokens_break tokens) el []) vl vll
+                 | ValFun _
+                 | ValPrim _
+                 | ValRules _
+                 | ValEnv _
+                 | ValBody _
+                 | ValMap _
+                 | ValObject _
+                 | ValChannel _
+                 | ValClass _
+                 | ValCases _
+                 | ValOther _
+                 | ValFunValue _
+                 | ValKey _ ->
+                      collect (Lm_string_util.tokens_atomic tokens (TokString v)) vl vll
+                 | ValApply _
+                 | ValImplicit _
+                 | ValSuperApply _
+                 | ValMethodApply _ ->
+                      raise (OmakeException (pos, StringValueError ("illegal application", v))))
+       | [], vl :: vll ->
+            collect tokens vl vll
+       | [], [] ->
+            Lm_string_util.tokens_flush tokens
+   in
+      collect tokens [v] []
+
+(*
+ * Flatten the value list into a arg_string list.
+ * Basically just concatenate all the values, being
+ * careful to preserve quoting.  In addition, we want to
+ * concatenate adjacent strings of the same type.
+ *)
+and arg_of_values venv pos vl =
+   let pos = string_pos "arg_of_values" pos in
+
+   (*
+    * Flatten all sequences.
+    *)
+   let rec collect is_quoted tokens vl vll =
+      match vl, vll with
+         v :: vl, _ ->
+            let v = eval_value venv pos v in
+               (match v with
+                   ValNone ->
+                      collect is_quoted tokens vl vll
+
+                   (* Strings *)
+                 | ValString s ->
+                      let tokens =
+                         if is_quoted then
+                            arg_buffer_add_data tokens s
+                         else
+                            arg_buffer_add_string tokens s
+                      in
+                         collect is_quoted tokens vl vll
+                 | ValData s ->
+                      collect is_quoted (arg_buffer_add_data tokens s) vl vll
+                 | ValSequence el ->
+                      collect is_quoted tokens el (vl :: vll)
+                 | ValArray el ->
+                      collect true tokens el (vl :: vll)
+
+                   (* Other quoted values *)
+                 | ValInt _
+                 | ValFloat _
+                 | ValDir _
+                 | ValNode _
+                 | ValQuote _
+                 | ValQuoteString _
+                 | ValFun _
+                 | ValPrim _
+                 | ValRules _
+                 | ValEnv _
+                 | ValBody _
+                 | ValMap _
+                 | ValObject _
+                 | ValChannel _
+                 | ValClass _
+                 | ValCases _
+                 | ValOther _
+                 | ValFunValue _
+                 | ValKey _ ->
+                      let tokens = arg_buffer_add_data tokens (string_of_value venv pos v) in
+                         collect is_quoted tokens vl vll
+
+                   (* Illegal values *)
+                 | ValApply _
+                 | ValImplicit _
+                 | ValSuperApply _
+                 | ValMethodApply _ ->
+                      raise (OmakeException (pos, StringValueError ("illegal application", v))))
+       | [], vl :: vll ->
+            collect is_quoted tokens vl vll
+       | [], [] ->
+            arg_buffer_contents tokens
+   in
+      collect false arg_buffer_empty vl []
+
+and argv_of_values venv pos vll =
+   List.map (arg_of_values venv pos) vll
+
 (*
  * Boolean test.
  * Arrays are always true.
@@ -694,7 +921,7 @@ and bool_of_value venv pos v =
        | [ValString s] ->
             Omake_util.bool_of_string s
        | [ValQuote vl] ->
-            Omake_util.bool_of_string (string_of_quote venv pos vl)
+            Omake_util.bool_of_string (string_of_quote venv pos None vl)
        | _ ->
             true
 
@@ -713,6 +940,7 @@ and file_of_value venv pos file =
        | ValString _
        | ValSequence _
        | ValQuote _
+       | ValQuoteString _
        | ValInt _
        | ValFloat _ ->
             venv_intern venv PhonyExplicit (string_of_value venv pos file)
@@ -749,19 +977,27 @@ and eval_value_core venv pos v =
       ValKey (loc, v) ->
          eval_key venv pos loc v
     | ValApply (loc, scope, v, []) ->
-         eval_value_core venv pos (eval_var venv pos loc (venv_find_var venv scope pos loc v))
+         begin match venv_find_var venv scope pos loc v with
+            ValApply (_, scope', v', []) when scope = scope' && Lm_symbol.eq v v' ->
+               let print_error buf =
+                  fprintf buf "@[<v0>@[<hv3>Self-referential variable:@ %a@]@ @[<hov3>The value of the variable@ %a%a@ is a lazy self-reference@ $`(%a%a),@ resulting in an infinite self-reference loop.@]@]" (**)
+                     pp_print_symbol v pp_print_scope_kind scope pp_print_symbol v pp_print_scope_kind scope pp_print_symbol v;
+               in
+                  raise (OmakeException(loc_pos loc pos, LazyError print_error))
+          | v ->
+               eval_value_core venv pos (eval_var venv pos loc v)
+         end
     | ValApply (loc, scope, v, args) ->
          eval_value_core venv pos (eval_apply venv pos loc (venv_find_var venv scope pos loc v) args)
     | ValImplicit (loc, scope, v) ->
-         begin match
-            try
-               Some (venv_find_var_exn venv scope v)
-            with Not_found ->
-               None
-         with
-            Some v -> ValArray [eval_value_core venv pos (eval_var venv pos loc v)]
-          | None -> ValNone
-         end
+         let v =
+            try Some (venv_find_var_exn venv scope v) with
+               Not_found ->
+                  None
+         in
+            (match v with
+                Some v -> ValArray [eval_value_core venv pos (eval_var venv pos loc v)]
+              | None -> ValNone)
     | ValSuperApply (loc, scope, super, v, args) ->
          let obj = venv_find_super venv pos loc super in
          let v = venv_find_field obj pos v in
@@ -849,6 +1085,7 @@ and eval_body_value venv pos v =
     | ValData _
     | ValString _
     | ValQuote _
+    | ValQuoteString _
     | ValDir _
     | ValNode _
     | ValFun _
@@ -870,6 +1107,44 @@ and eval_body_value venv pos v =
     | ValMethodApply _ ->
          raise (Invalid_argument "eval_body_value")
 
+(* XXX: JYH: this is temporary, there is no need for it in 0.9.9 *)
+and eval_body_value_env venv pos v =
+   match eval_value venv pos v with
+      ValSequence sl ->
+         venv, ValSequence (List.map (eval_body_value venv pos) sl)
+    | ValArray sl ->
+         venv, ValArray (List.map (eval_body_value venv pos) sl)
+    | ValBody (env, body) ->
+         let venv = venv_with_env venv env in
+            eval_exp venv ValNone body
+    | ValNone
+    | ValInt _
+    | ValFloat _
+    | ValData _
+    | ValString _
+    | ValQuote _
+    | ValQuoteString _
+    | ValDir _
+    | ValNode _
+    | ValFun _
+    | ValFunValue _
+    | ValPrim _
+    | ValRules _
+    | ValMap _
+    | ValObject _
+    | ValEnv _
+    | ValChannel _
+    | ValClass _
+    | ValCases _
+    | ValOther _ as result ->
+         venv, result
+    | ValKey _
+    | ValApply _
+    | ValImplicit _
+    | ValSuperApply _
+    | ValMethodApply _ ->
+         raise (Invalid_argument "eval_body_value_env")
+
 and eval_body_exp venv pos x v =
    match eval_value venv pos v with
       ValSequence sl ->
@@ -883,6 +1158,7 @@ and eval_body_exp venv pos x v =
     | ValFloat _
     | ValData _
     | ValQuote _
+    | ValQuoteString _
     | ValString _
     | ValDir _
     | ValNode _
@@ -1050,7 +1326,8 @@ and eval_object_exn venv pos x =
     | ValFloat _ ->
          create_object venv x float_object_sym
     | ValData _
-    | ValQuote _ ->
+    | ValQuote _
+    | ValQuoteString _ ->
          create_object venv x string_object_sym
     | ValSequence _
     | ValString _
@@ -1149,11 +1426,15 @@ and eval_method_var venv scope pos loc vl =
  * Simplify a quoted value if possible.
  * Strings are concatenated.
  *)
-and simplify_quote_val venv pos el =
+and simplify_quote_val venv pos c el =
    match el with
       [ValString s]
     | [ValData s] ->
-         ValData s
+         (match c with
+             None ->
+                ValData s
+           | Some c ->
+                ValQuoteString (c, [ValData s]))
     | _ ->
          let buf = Buffer.create 32 in
          let flush vl =
@@ -1182,7 +1463,11 @@ and simplify_quote_val venv pos el =
              | [ValData s] ->
                   ValData s
              | el ->
-                  ValQuote el
+                  match c with
+                     None ->
+                        ValQuote el
+                   | Some c ->
+                        ValQuoteString (c, el)
 
 (*
  * Evaluate a string expression.
@@ -1246,7 +1531,9 @@ and eval_string_exp be_eager venv pos s =
             in
                ValCases cases
        | QuoteString (_, el) ->
-            simplify_quote_val venv pos (List.map (eval_string_exp be_eager venv pos) el)
+            simplify_quote_val venv pos None (List.map (eval_string_exp be_eager venv pos) el)
+       | QuoteStringString (_, c, el) ->
+            simplify_quote_val venv pos (Some c) (List.map (eval_string_exp be_eager venv pos) el)
        | ThisString (_, scope) ->
             ValObject (venv_this_object venv scope)
 
@@ -1327,10 +1614,16 @@ and eval_let_var_exp venv pos scope v flag s =
             s
        | VarDefAppend ->
             let v = venv_get_var venv scope pos v in
-               if is_array_value v || is_array_value s then
+               (* Be lazy about concatenating arrays to avoid quadratic behavior *)
+               if is_array_value v then
+                  if is_array_value s then
+                     ValArray [v; s]
+                  else
+                     let sl = values_of_value venv pos s in
+                        ValArray (v :: sl)
+               else if is_array_value s then
                   let vl = values_of_value venv pos v in
-                  let sl = values_of_value venv pos s in
-                     ValArray (vl @ sl)
+                     ValArray [ValArray vl; s]
                else if is_empty_value v then
                   s
                else if is_empty_value s then
@@ -1558,25 +1851,27 @@ and eval_method_apply_exp venv scope pos loc vl args =
 and eval_export_exp venv pos s =
    let pos = string_pos "eval_export_exp" pos in
    let result = eager_string_exp venv pos s in
-   let result =
-      match result with
-         ValEnv _ ->
-            result
-       | _ ->
-            let args = strings_of_value venv pos result in
-            let syms =
-               match args with
-                  []
-                | ["all"] ->
-                     ExportAll
-                | ["rules"] ->
-                     ExportRules
-                | _ ->
-                     ExportSymbols (List.map Lm_symbol.add args)
-            in
-               ValEnv (venv, syms)
-   in
-      venv, result
+      venv, (eval_export_args venv pos result)
+
+and eval_export_args venv pos = function
+   ValEnv _ as result ->
+      result
+ | result ->
+      let args = strings_of_value venv pos result in
+      let syms =
+         match args with
+            [] ->
+               ExportAll
+          | ["all"] ->
+               eprintf "@[<hv3>*** omake WARNING: %a: @ @[<hov0>\"export all\" syntax is deprecated;@ use an empty \"export\" instead.@]@]@." pp_print_pos pos;
+               ExportAll
+          | ["rules"] ->
+               eprintf "@[<hv3>*** omake WARNING: %a:@ @[<hov0>\"export rules\" syntax is deprecated;@ use \"export .RULE .PHONY\" instead.@]@]@." pp_print_pos pos;
+               ExportList [ExportRules; ExportPhonies]
+          | _ ->
+               ExportList (List.map (eval_export_elt pos) args)
+      in
+         ValEnv (venv, syms)
 
 and eval_cancel_export_exp venv pos s =
    let v =
@@ -1635,7 +1930,7 @@ and include_file venv scope pos loc target =
  *)
 and eval_object_file venv pos loc node =
    let parse_obj info node =
-      let names, _, ir = compile_ir_info venv IncludePervasives info pos loc node in
+      let names, _, ir = compile_add_ir_info venv IncludePervasives info pos loc node in
       let venv = venv_get_pervasives venv node in
       let venv = venv_define_object venv in
       let venv, result = eval_exp venv ValNone ir in
@@ -1656,7 +1951,13 @@ let eval_open_file = open_ir
  * Project compiler.
  *)
 let compile venv =
-   let node = venv_intern venv PhonyProhibited makeroot_name in
+   let rootname =
+      if Sys.file_exists makeroot_name then
+         makeroot_name
+      else
+         makeroot_short_name
+   in
+   let node = venv_intern venv PhonyProhibited rootname in
    let venv = venv_add_file venv node in
    let loc = bogus_loc (Node.fullname node) in
    let pos = string_pos "compile" (loc_exp_pos loc) in
@@ -1681,12 +1982,9 @@ let compile_deps venv node buf =
             let sources = strings_of_value venv pos source in
                targets, sources) deps
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
