@@ -4,24 +4,28 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2007 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified by: Aleksey Nogin @email{nogin@metaprl.org}
  * @end[license]
  *)
 open Lm_glob
@@ -32,6 +36,7 @@ open Omake_env
 open Omake_node
 open Omake_ir_ast
 open Omake_exec_type
+open Omake_command_type
 
 val print_ast   : bool ref
 val print_ir    : bool ref
@@ -89,6 +94,7 @@ val eval_value : venv -> pos -> value -> value
  * Evaluate ValBody expressions.
  *)
 val eval_body_value : venv -> pos -> value -> value
+val eval_body_value_env : venv -> pos -> value -> venv * value
 val eval_body_exp   : venv -> pos -> value -> value -> venv * value
 
 (*
@@ -123,19 +129,27 @@ val eval_prim_value : venv -> pos -> value -> value
  *)
 val eval_apply : venv -> pos -> loc -> value -> value list -> value
 
+(*
+ * Evaluate "export" arguments.
+ *)
+val eval_export_args : venv -> pos -> value -> value
+
 (*
  * Conversions.
  * The following two functions should be used with care, since
  * they fail if the value contains an array.
  *)
 val string_of_value : venv -> pos -> value -> string
-val string_of_quote : venv -> pos -> value list -> string
+val string_of_quote : venv -> pos -> char option -> value list -> string
 val file_of_value   : venv -> pos -> value -> Node.t
 val path_of_values  : venv -> pos -> value list -> string -> (bool * Dir.t list) list
 
 (*
  * These conversions are safe to use anywhere.
  *)
+val tokens_of_value  : venv -> pos -> lexer -> value -> tok list
+val arg_of_values    : venv -> pos -> value list -> arg
+val argv_of_values   : venv -> pos -> value list list -> arg list
 val values_of_value  : venv -> pos -> value -> value list
 val strings_of_value : venv -> pos -> value -> string list
 val bool_of_value    : venv -> pos -> value -> bool
