@@ -8,8 +8,8 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -18,12 +18,17 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
  * @end[license]
  *)
+open Lm_glob
 open Lm_location
 
 open Omake_ir
@@ -70,6 +75,7 @@ val int_of_value        : venv -> pos -> value -> int
 val float_of_value      : venv -> pos -> value -> float
 val number_of_value     : venv -> pos -> value -> value
 val key_of_value        : venv -> pos -> value -> value
+val map_of_value        : venv -> pos -> value -> map
 
 val dir_of_value        : venv -> pos -> value -> Dir.t
 val file_of_value       : venv -> pos -> value -> Node.t
@@ -84,6 +90,9 @@ val channel_of_value         : venv -> pos -> value -> Lm_channel.t
 val in_channel_of_any_value  : venv -> pos -> value -> prim_channel * bool
 val out_channel_of_any_value : venv -> pos -> value -> prim_channel * bool
 
+val is_glob_value            : glob_options -> value -> bool
+val is_glob_value_list       : glob_options -> value list -> bool
+
 val current_lexer            : venv -> pos -> Lexer.t
 val current_parser           : venv -> pos -> Parser.t
 val loc_of_value             : venv -> pos -> value -> loc
