@@ -13,12 +13,12 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2007 Mojave Group, Caltech and HRL Laboratories, LLC
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -27,10 +27,14 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{anogin@hrl.com}
  * @end[license]
  *)
 open Lm_debug
@@ -77,7 +81,7 @@ type dir_listing = dir_listing_item list
  *)
 type exe_entry_core =
    ExeEntryCore of (Dir.t * string) list
- | ExeEntryNode of Node.t
+ | ExeEntryNodes of Node.t list
 
 type exe_listing_item = exe_entry_core ref StringTable.t
 
@@ -149,17 +153,13 @@ type cache =
    { (* State *)
      mutable cache_nodes             : node_memo NodeTable.t;
      mutable cache_info              : cache_info array;
-     mutable cache_file_stat_count   : int;  (* only succeeded stats are counted *)
+     mutable cache_file_stat_count   : int;  (* only succeessful stats are counted *)
      mutable cache_digest_count      : int;
 
      (* Path lookups *)
-     mutable cache_dirs         : (stat option * dir_listing_item) DirHashTable.t;
+     mutable cache_dirs         : (stat option * dir_listing_item) DirTable.t;
      mutable cache_path         : (stat option list * dir_listing_item) DirListTable.t;
-     mutable cache_exe_path     : (stat option list * exe_listing_item) DirListTable.t;
-
-     (* Hash-cons versions of directories and paths *)
-     cache_dir_hash             : DirHash.state;
-     cache_path_hash            : DirListHash.state
+     mutable cache_exe_path     : (stat option list * exe_listing_item) DirListTable.t
    }
 
 (*
@@ -167,8 +167,7 @@ type cache =
  * contains a table of all nodes.
  *)
 type cache_save =
-   { save_nodes              : Node.db;
-     save_cache_nodes        : node_memo NodeTable.t;
+   { save_cache_nodes        : node_memo NodeTable.t;
      save_cache_info         : memo NodeTable.t array
    }
 
@@ -190,7 +189,7 @@ let rule_fun            = 1
 let env_fun             = 2
 let include_fun         = 3
 
-let env_target  = Node.phony_global ".ENV"
+let env_target  = Node.create_phony_global ".ENV"
 
 (************************************************************************
  * Printing.
@@ -267,15 +266,13 @@ let create () =
      cache_info            = [||];
      cache_file_stat_count = 0;
      cache_digest_count    = 0;
-     cache_dirs            = DirHashTable.empty;
+     cache_dirs            = DirTable.empty;
      cache_path            = DirListTable.empty;
-     cache_exe_path        = DirListTable.empty;
-     cache_dir_hash        = DirHash.create_state ();
-     cache_path_hash       = DirListHash.create_state ()
+     cache_exe_path        = DirListTable.empty
    }
 
 let rehash cache =
-   cache.cache_dirs <- DirHashTable.empty;
+   cache.cache_dirs <- DirTable.empty;
    cache.cache_path <- DirListTable.empty;
    cache.cache_exe_path <- DirListTable.empty
 
@@ -303,7 +300,6 @@ let save_of_cache cache =
          cache_info  = cache_info
        } = cache
    in
-   let nodes = Node.marshal_base () in
    let cache_nodes =
       NodeTable.fold (fun nodes target nmemo ->
             match nmemo.nmemo_stats with
@@ -316,8 +312,7 @@ let save_of_cache cache =
                   nodes) NodeTable.empty cache_nodes
    in
    let cache_info = Array.map (fun info -> info.cache_memos) cache_info in
-      { save_nodes       = nodes;
-        save_cache_nodes = cache_nodes;
+      { save_cache_nodes = cache_nodes;
         save_cache_info  = cache_info
       }
 
@@ -343,8 +338,7 @@ let create_index memos =
  * Rebuild the cache from the saved version.
  *)
 let cache_of_save save =
-   let { save_nodes = nodes;
-         save_cache_nodes = cache_nodes;
+   let { save_cache_nodes = cache_nodes;
          save_cache_info  = cache_info
        } = save
    in
@@ -355,10 +349,8 @@ let cache_of_save save =
                  cache_index = index
                }) cache_info
    in
-      Node.unmarshal_base nodes;
-      { (create ()) with
-        cache_nodes = cache_nodes;
-        cache_info  = cache_info;
+      { (create ()) with cache_nodes = cache_nodes;
+                         cache_info  = cache_info;
       }
 
 (*
@@ -440,6 +432,69 @@ let pp_print_stat buf stat =
    in
       fprintf buf "ino = %d, size = %Ld, mtime = %g" ino size mtime
 
+(*
+ * Sloppy digests on files larger than 16MB.
+ *)
+let large_file_size = 8_000_000L
+let sample_size     = 0x10000
+let sample_count    = 16
+
+let rec really_read_exn fd buf off len =
+   if len > 0 then
+      let amount = Unix.read fd buf off len in
+         really_read_exn fd buf (off + amount) (len - amount)
+
+let digest_sample_exn fd file_size =
+   let sample = String.create (8 + (sample_count + 1) * sample_size) in
+      (* Add the file length to the string *)
+      sample.[0] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 56));
+      sample.[1] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 48));
+      sample.[2] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 40));
+      sample.[3] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 32));
+      sample.[4] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 24));
+      sample.[5] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 16));
+      sample.[6] <- Char.unsafe_chr (Int64.to_int (Int64.shift_right_logical file_size 8));
+      sample.[7] <- Char.unsafe_chr (Int64.to_int file_size);
+
+      (* Sample the file *)
+      for i = 0 to sample_count do
+         let off =
+            if i = sample_count then
+               Int64.sub file_size (Int64.of_int sample_size)
+            else
+               Int64.div (Int64.mul file_size (Int64.of_int i)) (Int64.of_int sample_count)
+         in
+         let _ = Unix.LargeFile.lseek fd off Unix.SEEK_SET in
+            really_read_exn fd sample (8 + i * sample_size) sample_size
+      done;
+
+      (* Take the digest of the sample *)
+      Digest.string sample
+
+let digest_large name file_size =
+   try
+      let fd = Unix.openfile name [Unix.O_RDONLY] 0o000 in
+      let digest =
+         try
+            digest_sample_exn fd file_size
+         with
+            Unix.Unix_error _
+          | End_of_file ->
+               Unix.close fd;
+               raise (Sys_error name)
+      in
+         Unix.close fd;
+         digest
+   with
+      Unix.Unix_error _ ->
+         raise (Sys_error name)
+
+let digest_file name file_size =
+   if file_size > large_file_size then
+      digest_large name file_size
+   else
+      Digest.file name
+
 (*
  * Stat a file.
  *)
@@ -480,7 +535,7 @@ let stat_file cache node =
                         if stats_equal stats' stats then
                            { nmemo with nmemo_stats = FreshStats stats' }
                         else
-                           let digest = Digest.file name in
+                           let digest = digest_file name stats'.Unix.LargeFile.st_size in
                               cache.cache_digest_count <- succ cache.cache_digest_count;
                               { nmemo_stats = FreshStats stats';
                                 nmemo_digest = Some digest
@@ -509,7 +564,7 @@ let stat_file cache node =
                        nmemo_digest = squash_stat
                      }
                   else
-                     let digest = Digest.file name in
+                     let digest = digest_file name stats.Unix.LargeFile.st_size in
                         cache.cache_digest_count <- succ cache.cache_digest_count;
                         { nmemo_stats = FreshStats stats;
                           nmemo_digest = Some digest
@@ -539,7 +594,7 @@ let stat_file cache node =
                         }
                      else begin
                         cache.cache_file_stat_count <- succ cache.cache_file_stat_count;
-                        let digest = Digest.file name in
+                        let digest = digest_file name stats.Unix.LargeFile.st_size in
                            cache.cache_digest_count <- succ cache.cache_digest_count;
                            { nmemo_stats = FreshStats stats;
                              nmemo_digest = Some digest
@@ -986,21 +1041,19 @@ let rec list_directory cache dir =
  * Get the directory listing as a StringTable.
  *)
 let ls_dir cache auto_rehash dir =
-   let key = DirHash.create cache.cache_dir_hash dir in
    let stat = lazy (stat_dir cache dir) in
-      try check_stat auto_rehash (DirHashTable.find cache.cache_dirs key) stat with
+      try check_stat auto_rehash (DirTable.find cache.cache_dirs dir) stat with
          Not_found ->
             let entries = list_directory cache dir in
             let stat = Lazy.force stat in
-               cache.cache_dirs <- DirHashTable.add cache.cache_dirs key (stat, entries);
+               cache.cache_dirs <- DirTable.add cache.cache_dirs dir (stat, entries);
                entries
 
 (*
  * Path version.
  *)
 let ls_path cache auto_rehash dirs =
-   let keys = List.map (DirHash.create cache.cache_dir_hash) dirs in
-   let key = DirListHash.create cache.cache_path_hash keys in
+   let key = DirListHash.create dirs in
    let stats = lazy (stat_dirs cache dirs) in
       try check_stats auto_rehash (DirListTable.find cache.cache_path key) stats with
          Not_found ->
@@ -1027,7 +1080,7 @@ let listing_find_item cache listing s =
          DirEntryCore entry ->
             entry
        | LazyEntryCore (dir, s) ->
-            let node = Node.intern no_mount_points PhonyProhibited dir s in
+            let node = Node.create_node no_mount_info Mount.empty dir s in
             let entry =
                if is_dir cache node then
                   DirEntry (Dir.chdir dir s)
@@ -1043,19 +1096,20 @@ let listing_find_item cache listing s =
  *
  * On Win32:
  *    - File permission doesn't matter
- *    - Files without suffix, and with .exe and .bat suffixes are executable
+ *    - Files without suffix, and with .com, .exe, .bat and .cmd suffixes are executable
  *
  * On Cygwin:
  *    - Files without suffix must be executable
- *    - Files with .exe and .bat suffixes are executable
+ *    - Files with .com, .exe, .bat and .cmd suffixes are executable
  *
  * On Unix:
  *    - There is no .exe suffix
  *    - Only files that are executable count
  *)
+let win32_suffixes = [".com"; ".exe"; ".bat"; ".cmd"]
+
 let ls_exe_path_win32 cache auto_rehash dirs =
-   let keys = List.map (DirHash.create cache.cache_dir_hash) dirs in
-   let key = DirListHash.create cache.cache_path_hash keys in
+   let key = DirListHash.create dirs in
    let stats = lazy (stat_dirs cache dirs) in
       try check_stats auto_rehash (DirListTable.find cache.cache_exe_path key) stats with
          Not_found ->
@@ -1066,12 +1120,12 @@ let ls_exe_path_win32 cache auto_rehash dirs =
                            StringTable.fold (fun entries name _ ->
                                  let info = dir, name in
                                  let name = String.lowercase name in
-                                    if Filename.check_suffix name ".exe" || Filename.check_suffix name ".bat" then
-                                       let name_core = Filename.chop_extension name in
-                                       let entries = StringMTable.add entries name_core info in
+                                 let entries = StringMTable.add entries name info in
+                                    if List.exists (Filename.check_suffix name) win32_suffixes then
+                                       let name = Filename.chop_extension name in
                                           StringMTable.add entries name info
                                     else
-                                       StringMTable.add entries name info) entries1 entries2
+                                       entries) entries1 entries2
                      with
                         Not_found ->
                            entries1) StringMTable.empty (List.rev dirs)
@@ -1085,8 +1139,7 @@ let ls_exe_path_win32 cache auto_rehash dirs =
                entries
 
 let ls_exe_path_unix cache auto_rehash dirs =
-   let keys = List.map (DirHash.create cache.cache_dir_hash) dirs in
-   let key = DirListHash.create cache.cache_path_hash keys in
+   let key = DirListHash.create dirs in
    let stats = lazy (stat_dirs cache dirs) in
       try check_stats auto_rehash (DirListTable.find cache.cache_exe_path key) stats with
          Not_found ->
@@ -1117,7 +1170,7 @@ let is_exe_file cache node =
             Unix.LargeFile.st_perm = perm;
             Unix.LargeFile.st_uid = uid;
             Unix.LargeFile.st_gid = gid
-          } = Unix.LargeFile.stat (Node.fullname node)
+          } = stat_unix cache node
       in
          (kind = Unix.S_REG)
          && ((perm land 0o001) <> 0
@@ -1128,56 +1181,82 @@ let is_exe_file cache node =
     | Not_found ->
          false
 
-let rec search_exe_win32 cache entries =
-   match entries with
-      (dir, s) :: entries ->
-         let node = Node.intern no_mount_points PhonyProhibited dir s in
-            if is_dir cache node then
-               search_exe_win32 cache entries
-            else
-               node
-    | [] ->
-         raise Not_found
+let is_exe_win32 cache dir s =
+   let node = Node.create_node no_mount_info Mount.empty dir s in
+      if is_dir cache node then
+         None
+      else
+         Some node
 
-let rec search_exe_unix cache entries =
-   match entries with
-      (dir, s) :: entries ->
-         let node = Node.intern no_mount_points PhonyProhibited dir s in
-            if is_exe_file cache node then
-               node
-            else
-               search_exe_unix cache entries
-    | [] ->
-         raise Not_found
+let is_exe_unix cache dir s =
+   let node = Node.create_node no_mount_info Mount.empty dir s in
+      if is_exe_file cache node then
+         Some node
+      else
+         None
 
-let rec search_exe_cygwin cache entries =
-   match entries with
-      (dir, s) :: entries ->
-         let node = Node.intern no_mount_points PhonyProhibited dir s in
-            if Filename.check_suffix s ".exe" || Filename.check_suffix s ".bat" || is_exe_file cache node then
-               node
-            else
-               search_exe_cygwin cache entries
-    | [] ->
-         raise Not_found
+let is_exe_cygwin cache dir s =
+   let node = Node.create_node no_mount_info Mount.empty dir s in
+      if List.exists (Filename.check_suffix s) win32_suffixes || is_exe_file cache node then
+         Some node
+      else
+         None
 
-let ls_exe_path, search_exe, name_exe =
+let ls_exe_path, is_exe, name_exe, exe_suffixes =
    if Sys.os_type = "Win32" then
-      ls_exe_path_win32, search_exe_win32, String.lowercase
+      ls_exe_path_win32, is_exe_win32, String.lowercase, "" :: win32_suffixes
    else if Sys.os_type = "Cygwin" then
-      ls_exe_path_win32, search_exe_cygwin, String.lowercase
+      ls_exe_path_win32, is_exe_cygwin, String.lowercase, "" :: win32_suffixes
    else
-      ls_exe_path_unix, search_exe_unix, (fun s -> s)
+      ls_exe_path_unix, is_exe_unix, (fun s -> s), [""]
 
-let exe_find_item cache listing s =
+let rec search_exe cache entries =
+   Lm_list_util.some_map (fun (dir, s) ->
+         is_exe cache dir s) entries
+
+let exe_find_nodes cache listing s =
    let entry_ref = StringTable.find listing (name_exe s) in
       match !entry_ref with
          ExeEntryCore entries ->
-            let node = search_exe cache entries in
-               entry_ref := ExeEntryNode node;
-               node
-       | ExeEntryNode node ->
-            node
+            let nodes = search_exe cache entries in
+               entry_ref := ExeEntryNodes nodes;
+               nodes
+       | ExeEntryNodes nodes ->
+            nodes
+
+let exe_find_nodes_all cache listing s =
+   try exe_find_nodes cache listing s with
+      Not_found ->
+         []
+
+let exe_find_item cache listing s =
+   match exe_find_nodes cache listing s with
+      node :: _ ->
+         node
+    | [] ->
+         raise Not_found
+
+(*
+ * Find all entries with the given prefix.
+ *)
+let exe_complete_prefix cache s items listing =
+   StringTable.fold (fun items s2 entry_ref ->
+         if Lm_string_util.is_string_prefix s s2 then
+            let nodes =
+               match !entry_ref with
+                  ExeEntryCore entries ->
+                     let nodes = search_exe cache entries in
+                        entry_ref := ExeEntryNodes nodes;
+                        nodes
+                | ExeEntryNodes nodes ->
+                     nodes
+            in
+               if nodes = [] then
+                  items
+               else
+                  StringSet.add items s2
+         else
+            items) items listing
 
 (************************************************************************
  * Redefine the functions to work on directory groups, where each group may
@@ -1205,6 +1284,12 @@ let rec exe_find cache listings s =
     | [] ->
          raise Not_found
 
+let exe_find_all cache listings s =
+   List.flatten (List.map (fun listing -> exe_find_nodes_all cache listing s) listings)
+
+let exe_complete cache listings s =
+   List.fold_left (exe_complete_prefix cache s) StringSet.empty listings
+
 let ls_dir cache auto_rehash dir =
    [ls_dir cache auto_rehash dir]
 
@@ -1216,12 +1301,9 @@ let ls_exe_path cache groups =
    List.map (fun (auto_rehash, dirs) ->
          ls_exe_path cache auto_rehash dirs) groups
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
