@@ -10,8 +10,8 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -20,7 +20,11 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -52,6 +56,8 @@ let shell_object_sym           = Lm_symbol.add "Shell"
 let select_object_sym          = Lm_symbol.add "Select"
 let pipe_object_sym            = Lm_symbol.add "Pipe"
 let stat_object_sym            = Lm_symbol.add "Stat"
+let passwd_object_sym          = Lm_symbol.add "Passwd"
+let group_object_sym           = Lm_symbol.add "Group"
 let lexer_object_sym           = Lm_symbol.add "Lexer"
 let parser_object_sym          = Lm_symbol.add "Parser"
 let location_object_sym        = Lm_symbol.add "Location"
@@ -75,14 +81,27 @@ let st_ino_sym                 = Lm_symbol.add "ino"
 let st_kind_sym                = Lm_symbol.add "kind"
 let st_perm_sym                = Lm_symbol.add "perm"
 let st_nlink_sym               = Lm_symbol.add "nlink"
-let st_uid_sym                 = Lm_symbol.add "uid_sym  "
-let st_gid_sym                 = Lm_symbol.add "gid_sym  "
-let st_rdev_sym                = Lm_symbol.add "rdev_sym "
-let st_size_sym                = Lm_symbol.add "size_sym "
+let st_uid_sym                 = Lm_symbol.add "uid"
+let st_gid_sym                 = Lm_symbol.add "gid"
+let st_rdev_sym                = Lm_symbol.add "rdev"
+let st_size_sym                = Lm_symbol.add "size"
 let st_atime_sym               = Lm_symbol.add "atime"
 let st_mtime_sym               = Lm_symbol.add "mtime"
 let st_ctime_sym               = Lm_symbol.add "ctime"
 
+let pw_name_sym                = Lm_symbol.add "pw_name"
+let pw_passwd_sym              = Lm_symbol.add "pw_passwd"
+let pw_uid_sym                 = Lm_symbol.add "pw_uid"
+let pw_gid_sym                 = Lm_symbol.add "pw_gid"
+let pw_gecos_sym               = Lm_symbol.add "pw_gecos"
+let pw_dir_sym                 = Lm_symbol.add "pw_dir"
+let pw_shell_sym               = Lm_symbol.add "pw_shell"
+
+let gr_name_sym                = Lm_symbol.add "gr_name"
+let gr_passwd_sym              = Lm_symbol.add "gr_passwd"
+let gr_gid_sym                 = Lm_symbol.add "gr_gid"
+let gr_mem_sym                 = Lm_symbol.add "gr_mem"
+
 let target_sym                 = Lm_symbol.add "target"
 let target_effects_sym         = Lm_symbol.add "effects"
 let scanner_deps_sym           = Lm_symbol.add "scanner-deps"
@@ -90,6 +109,7 @@ let static_deps_sym            = Lm_symbol.add "static-deps"
 let build_deps_sym             = Lm_symbol.add "build-deps"
 let build_values_sym           = Lm_symbol.add "build-values"
 let build_commands_sym         = Lm_symbol.add "build-commands"
+let output_file_sym            = Lm_symbol.add "output-file"
 
 let argv_sym                   = Lm_symbol.add "argv"
 let star_sym                   = Lm_symbol.add "*"
@@ -102,6 +122,7 @@ let hat_sym                    = Lm_symbol.add "^"
 let zero_sym                   = Lm_symbol.add "0"
 
 let runtime_exception_sym      = Lm_symbol.add "RuntimeException"
+let unbuildable_exception_sym  = Lm_symbol.add "UnbuildableException"
 let loc_sym                    = Lm_symbol.add "loc"
 let pos_sym                    = Lm_symbol.add "position"
 let message_sym                = Lm_symbol.add "message"
@@ -132,6 +153,8 @@ let scanner_mode_sym               = Lm_symbol.add "SCANNER_MODE"
 let history_file_sym               = Lm_symbol.add "history-file"
 let history_length_sym             = Lm_symbol.add "history-length"
 
+let build_summary_sym              = Lm_symbol.add "BUILD_SUMMARY"
+
 (*
  * Special symbols.
  *)
@@ -142,6 +165,8 @@ let else_sym      = Lm_symbol.add "else"
 let elseif_sym    = Lm_symbol.add "elseif"
 let switch_sym    = Lm_symbol.add "switch"
 let case_sym      = Lm_symbol.add "case"
+let do_sym        = Lm_symbol.add "do"
+let while_sym     = Lm_symbol.add "while"
 let default_sym   = Lm_symbol.add "default"
 let include_sym   = Lm_symbol.add "include"
 let section_sym   = Lm_symbol.add "section"
@@ -152,15 +177,19 @@ let finally_sym   = Lm_symbol.add "finally"
 let private_sym   = Lm_symbol.add "private"
 let protected_sym = Lm_symbol.add "protected"
 let public_sym    = Lm_symbol.add "public"
-let static_sym    = Lm_symbol.add "static"
 let rule_sym      = Lm_symbol.add "rule"
 let system_sym    = Lm_symbol.add "system"
 let open_sym      = Lm_symbol.add "open"
+let autoload_sym  = Lm_symbol.add "autoload"
 let declare_sym   = Lm_symbol.add "declare"
 let return_sym    = Lm_symbol.add "return"
 let export_sym    = Lm_symbol.add "export"
 let value_sym     = Lm_symbol.add "value"
 
+let empty_map_sym       = Lm_symbol.add "empty-map"
+let create_map_sym      = Lm_symbol.add "create-map"
+let create_lazy_map_sym = Lm_symbol.add "create-lazy-map"
+
 (*
  * Awk values.
  *)
@@ -177,7 +206,7 @@ let fsubst_sym    = Lm_symbol.add "fsubst"
 let cases_syms    = [awk_sym; fsubst_sym]
 let cases_set     = List.fold_left SymbolSet.add SymbolSet.empty cases_syms
 
-let clauses_syms  = [case_sym; default_sym; when_sym; catch_sym; finally_sym]
+let clauses_syms  = [case_sym; default_sym; when_sym; catch_sym; finally_sym; do_sym]
 let clauses_set   = List.fold_left SymbolSet.add SymbolSet.empty clauses_syms
 
 (*
