@@ -4,33 +4,55 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2006 Mojave Group, Caltech
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation,
+ * version 2.1 of the License.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ * 
+ * Additional permission is given to link this library with the
+ * OpenSSL project's "OpenSSL" library, and with the OCaml runtime,
+ * and you may distribute the linked executables.  See the file
+ * LICENSE.libmojave for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
  * @end[license]
  */
 #include <stdio.h>
+#include <caml/signals.h>
 #include <caml/mlvalues.h>
 #include <caml/alloc.h>
 #include <caml/memory.h>
 #include <caml/custom.h>
 #include <caml/fail.h>
 
+#if defined(WIN32) || defined(_WIN32)
+/* Disable some of the warnings */
+#pragma warning( disable : 4100 4201 4127 4189 4702 4996 )
+#endif
+
+/*
+ * Lock codes.
+ */
+#define LM_LOCK_UN      0
+#define LM_LOCK_SH      1
+#define LM_LOCK_EX      2
+#define LM_LOCK_TSH     3
+#define LM_LOCK_TEX     5
+
+#define FLOCK_LEN       ((unsigned int) ~0 >> 2)
+
 /*
  * Print the stack pointer for debugging.
  */
@@ -87,8 +109,8 @@ value lockf_win32(value v_fd, value v_kind, value v_len)
     int kind = Int_val(v_kind);
     int len = Int_val(v_len);
     OVERLAPPED overlapped;
-    int code, flags;
-    DWORD pos, error;
+    int code, flags = 0;
+    DWORD pos, error = 0;
 
     /* Get the current position in the file */
     pos = SetFilePointer(fd, 0, 0, FILE_CURRENT);
@@ -173,6 +195,28 @@ value lockf_win32(value v_fd, value v_kind, value v_len)
     return Val_unit;
 }
 
+/*
+ * Translate flock operators.
+ */
+static int lockf_of_flock[] = {
+    F_ULOCK,
+    F_RLOCK,
+    F_LOCK,
+    F_TRLOCK,
+    F_TLOCK
+};
+
+/*
+ * flock wrapper.
+ */
+value lm_flock(value v_fd, value v_op)
+{
+    value v_kind;
+
+    v_kind = Val_int(lockf_of_flock[Int_val(v_op)]);
+    return lockf_win32(v_fd, v_kind, Val_int(FLOCK_LEN));
+}
+
 /*
  * Truncate to the current position.
  */
@@ -196,7 +240,7 @@ value caml_registry_find(value v_hkey, value v_subkey, value v_field)
     const char *subkey, *field;
     DWORD len;
     LONG code;
-    HKEY hkey;
+    HKEY hkey = 0;
 
     /* Get the arguments */
     switch(Int_val(v_hkey)) {
@@ -249,6 +293,11 @@ value caml_registry_find(value v_hkey, value v_subkey, value v_field)
 }
 
 #else /* WIN32 */
+#include <unistd.h>
+#include <fcntl.h>
+#include <sys/file.h>
+#include <sys/types.h>
+#include <pwd.h>
 
 value int_of_fd(value fd)
 {
@@ -279,4 +328,149 @@ value caml_registry_find(value v_key, value v_subkey, value v_field)
     return Val_unit;
 }
 
+/*
+ * Translations.
+ */
+#if defined(LOCK_UN) && defined(LOCK_SH) && defined(LOCK_EX)
+#define FLOCK_ENABLED
+static int flock_of_flock[] = {
+    LOCK_UN,
+    LOCK_SH,
+    LOCK_EX,
+    LOCK_SH | LOCK_NB,
+    LOCK_EX | LOCK_NB
+};
+#endif
+
+#if defined(F_RDLCK) && defined(F_WRLCK) && defined(F_UNLCK) && defined(F_SETLK) && defined(F_SETLKW) && defined(SEEK_SET)
+#define FCNTL_ENABLED
+static int fcntl_type_of_flock[] = {
+    F_UNLCK,
+    F_RDLCK,
+    F_WRLCK,
+    F_RDLCK,
+    F_WRLCK
+};
+
+static int fcntl_of_flock[] = {
+    F_SETLKW,
+    F_SETLKW,
+    F_SETLKW,
+    F_SETLK,
+    F_SETLK
+};
+#endif
+
+#if defined(F_ULOCK) && defined(F_LOCK) && defined(F_TLOCK)
+#define LOCKF_ENABLED
+static int lockf_of_flock[] = {
+    F_ULOCK,
+    F_LOCK,
+    F_LOCK,
+    F_TLOCK,
+    F_TLOCK
+};
+#endif
+
+value lm_flock(value v_fd, value v_op)
+{
+    int fd, op, cmd, code;
+
+    fd = Int_val(v_fd);
+    op = Int_val(v_op);
+#if defined(FLOCK_ENABLED)
+    cmd = flock_of_flock[op];
+    enter_blocking_section();
+    code = flock(fd, cmd);
+    leave_blocking_section();
+#elif defined(FCNTL_ENABLED)
+    {
+        struct flock info;
+        cmd = fcntl_of_flock[op];
+        info.l_type = fcntl_type_of_flock[op];
+        info.l_whence = SEEK_SET;
+        info.l_start = 0;
+        info.l_len = FLOCK_LEN;
+        enter_blocking_section();
+        code = fcntl(fd, cmd, &info);
+        leave_blocking_section();
+    }
+#elif defined(LOCKF_ENABLED)
+    cmd = lockf_of_flock[op];
+    enter_blocking_section();
+    code = lockf(fd, cmd, FLOCK_LEN);
+    leave_blocking_section();
+#else
+    code = -1;
+#endif
+    if(code < 0)
+        failwith("flock");
+    return Val_unit;
+}
+
 #endif /* !WIN32 */
+
+/************************************************************************
+ * Password file (only on Unix).
+ */
+#ifdef WIN32
+
+/*
+ * The empty array.
+ */
+value lm_getpwents(value v_unit)
+{
+    return Val_emptylist;
+}
+
+#else /* !WIN32 */
+
+/*
+ * Scan the password file.
+ *
+type passwd_entry = {
+  	pw_name : string;
+  	pw_passwd : string;
+  	pw_uid : int;
+  	pw_gid : int;
+  	pw_gecos : string;
+  	pw_dir : string;
+  	pw_shell : string;
+}
+ */
+value lm_getpwents(value v_unit)
+{
+    CAMLparam1(v_unit);
+    CAMLlocal3(users, entry, cons);
+    struct passwd *entryp;
+
+    /* Create a list of users */
+    users = Val_emptylist;
+
+    /* Scan the password file */
+    setpwent();
+    while((entryp = getpwent())) {
+        entry = caml_alloc_tuple(7);
+        Store_field(entry, 0, caml_copy_string(entryp->pw_name));
+        Store_field(entry, 1, caml_copy_string(entryp->pw_passwd));
+        Store_field(entry, 2, Val_int(entryp->pw_uid));
+        Store_field(entry, 3, Val_int(entryp->pw_gid));
+#ifdef __BEOS__
+        Store_field(entry, 4, copy_string(""));
+#else
+        Store_field(entry, 4, copy_string(entryp->pw_gecos));
+#endif
+        Store_field(entry, 5, copy_string(entryp->pw_dir));
+        Store_field(entry, 6, copy_string(entryp->pw_shell));
+        cons = caml_alloc_tuple(2);
+        Store_field(cons, 0, entry);
+        Store_field(cons, 1, users);
+        users = cons;
+    }
+    endpwent();
+
+    CAMLreturn(users);
+}
+
+#endif /* !WIN32 */
+   
