@@ -7,19 +7,24 @@
  * @begin[license]
  * Copyright (C) 2003 Mojave Group, Caltech
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation,
+ * version 2.1 of the License.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ * 
+ * Additional permission is given to link this library with the
+ * OpenSSL project's "OpenSSL" library, and with the OCaml runtime,
+ * and you may distribute the linked executables.  See the file
+ * LICENSE.libmojave for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
@@ -65,7 +70,7 @@ type thread_request =
 
 (*
  * A handle is a particular copy of a value.
- *    handle_value: the actula data value
+ *    handle_value: the actual data value
  *    handle_queue: the queue of waiting threads
  *    handle_state: the current active state
  *    handle_active: a table of the currently active threads
@@ -624,6 +629,7 @@ struct
    let create = create_thread
    let self = ThreadCore.self
    let id = ThreadCore.id
+   let sigmask = ThreadCore.sigmask
 end
 
 (*
