@@ -10,24 +10,30 @@
  * See the file doc/htmlman/default.html or visit http://metaprl.org/
  * for more information.
  *
- * Copyright (C) 1998-2005 Jason Hickey, Cornell University
+ * Copyright (C) 1998-2007 PRL Group, Cornell University, California
+ * Institute of Technology and HRL Laboratories, LLC
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
- * This program is distributed in the hope that it will be useful,
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation,
+ * version 2.1 of the License.
+ * 
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
+ * 
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
+ * 
+ * Additional permission is given to link this library with the
+ * OpenSSL project's "OpenSSL" library, and with the OCaml runtime,
+ * and you may distribute the linked executables.  See the file
+ * LICENSE.libmojave for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
- *
- * Author: Jason Hickey
- * jyh@cs.cornell.edu
+ * Author: Jason Hickey <jyh@cs.cornell.edu>
+ * Modified By: Aleksey Nogin <anogin@hrl.com>
  *)
 open Lm_printf
 
@@ -43,6 +49,7 @@ type registry_hkey =
 
 external print_stack_pointer : unit -> unit = "lm_print_stack_pointer"
 external registry_find   : registry_hkey -> string -> string -> string = "caml_registry_find"
+external getpwents : unit -> Unix.passwd_entry list = "lm_getpwents"
 
 (*
  * Read the exact amount.
@@ -76,21 +83,25 @@ let rec copy_file_fd buffer from_fd to_fd =
 
 let copy_file from_name to_name mode =
    let from_fd = Unix.openfile from_name [Unix.O_RDONLY] 0o666 in
+   let () =
       try
          let to_fd = Unix.openfile to_name [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC] 0o600 in
+         let () =
             try
                copy_file_fd (String.create 8192) from_fd to_fd;
-               Unix.close from_fd;
-               Unix.close to_fd;
-               Unix.chmod to_name mode
+               Unix.fchmod to_fd mode
             with
                x ->
                   Unix.close to_fd;
                   raise x
+         in
+            Unix.close to_fd
       with
          x ->
             Unix.close from_fd;
             raise x
+   in
+      Unix.close from_fd
 
 (*
  * Make a directory hierarchy.
@@ -195,6 +206,15 @@ let ftruncate =
    else
       (fun fd -> Unix.ftruncate fd (Unix.lseek fd 0 Unix.SEEK_CUR))
 
+type flock_command =
+   LOCK_UN
+ | LOCK_SH
+ | LOCK_EX
+ | LOCK_TSH
+ | LOCK_TEX
+
+external flock : Unix.file_descr -> flock_command -> unit = "lm_flock"
+
 (*
  * Open a file descriptor.
  * This hook is here so you can add print statements to
@@ -205,7 +225,6 @@ let openfile = Unix.openfile
 (*
  * -*-
  * Local Variables:
- * Caml-master: "refiner"
  * End:
  * -*-
  *)
