@@ -20,24 +20,30 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Mojave Group, Caltech
+ * Copyright (C) 2003-2007 Mojave Group, California Institute of Technology
+ * and HRL Laboratories, LLC
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation,
+ * version 2.1 of the License.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * OpenSSL project's "OpenSSL" library, and with the OCaml runtime,
+ * and you may distribute the linked executables.  See the file
+ * LICENSE.libmojave for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{anogin@hrl.com}
  * @end[license]
  *)
 open Lm_printf
@@ -66,6 +72,7 @@ type mode =
  * Canonical names.
  *)
 let string_sym = Lm_symbol.add ".string"
+let fun_sym    = Lm_symbol.add ".fun"
 
 (*
  * The channel has an input and output buffer.
@@ -73,7 +80,7 @@ let string_sym = Lm_symbol.add ".string"
  * to perform asynchronous writes on pipes.
  *)
 type channel =
-   { channel_id           : int;
+   { mutable channel_id   : int;
      channel_fd           : Unix.file_descr option;
      channel_kind         : kind;
      channel_mode         : mode;
@@ -114,9 +121,12 @@ type channel =
 
      (*
       * out_max is the total amount of data in the output buffer.
+      * If out_expand is true, the outbuffer is expanded instead
+      * instead of flushing.
       *)
+     out_expand           : bool;
      mutable out_max      : int;
-     out_buffer           : string;
+     mutable out_buffer   : string;
 
      (*
       * In text mode, the output is double-buffered because
@@ -159,13 +169,16 @@ external pipe_kind : Unix.file_descr -> kind = "omake_shell_pipe_kind"
 
 (*
  * Default readers and writers.
+ * 
+ * XXX: We treat the "broken pipe" errors as EOFs.
  *)
 let default_reader fd buf off len =
    match fd with
       Some fd ->
          blocking_section (fun () ->
                try Unix.read fd buf off len with
-                  Unix.Unix_error (Unix.EUNKNOWNERR 0, _, _) ->
+                  Unix.Unix_error (Unix.EUNKNOWNERR 0, _, _)
+                | Unix.Unix_error (Unix.EPIPE, _, _) ->
                      0) ()
     | None ->
          raise (Unix.Unix_error (Unix.EINVAL, "default_reader", ""))
@@ -195,7 +208,7 @@ let null_writer _ _ _ =
 (*
  * Empty buffer.
  *)
-let create id file kind mode binary fd =
+let create file kind mode binary fd =
    let kind =
       match fd with
          Some fd ->
@@ -214,7 +227,7 @@ let create id file kind mode binary fd =
       else
          String.create (buf_size * 2)
    in
-      { channel_id     = id;
+      { channel_id     = 0;
         channel_fd     = fd;
         channel_kind   = kind;
         channel_mode   = mode;
@@ -233,6 +246,7 @@ let create id file kind mode binary fd =
         lex_index      = 0;
 
         out_max        = 0;
+        out_expand     = false;
         out_buffer     = out_buffer;
 
         write_pid      = 0;
@@ -244,6 +258,9 @@ let create id file kind mode binary fd =
         write_fun      = default_writer fd
       }
 
+let set_id info id =
+   info.channel_id <- id
+
 let of_string file line char s =
    let len = String.length s in
       { channel_id     = 0;
@@ -265,6 +282,7 @@ let of_string file line char s =
         lex_index    = 0;
 
         out_max      = 0;
+        out_expand   = false;
         out_buffer   = "";
 
         write_pid    = 0;
@@ -276,6 +294,38 @@ let of_string file line char s =
         write_fun    = null_writer
       }
 
+let of_fun read write =
+   { channel_id     = 0;
+     channel_fd     = None;
+     channel_kind   = FileChannel;
+     channel_mode   = InOutChannel;
+     channel_file   = fun_sym;
+     channel_binary = true;
+
+     start_line     = 1;
+     start_char     = 0;
+     middle_index   = 0;
+     middle_line    = 1;
+     middle_char    = 0;
+
+     in_index     = 0;
+     in_max       = 0;
+     in_buffer    = String.create buf_size;
+     lex_index    = 0;
+
+     out_max      = 0;
+     out_expand   = false;
+     out_buffer   = String.create buf_size;
+
+     write_pid    = 0;
+     write_index  = 0;
+     write_max    = 0;
+     write_buffer = String.create buf_size;
+
+     read_fun     = read;
+     write_fun    = write
+   }
+
 let of_loc_string file line char s =
    of_string (Lm_symbol.add file) line char s
 
@@ -314,6 +364,44 @@ let set_io_functions info reader writer =
    info.read_fun <- reader;
    info.write_fun <- writer
 
+let create_loc_string_aux file line char =
+   { channel_id     = 0;
+     channel_fd     = None;
+     channel_kind   = FileChannel;
+     channel_mode   = OutChannel;
+     channel_file   = file;
+     channel_binary = true;
+
+     start_line     = line;
+     start_char     = char;
+     middle_index   = 0;
+     middle_line    = line;
+     middle_char    = char;
+
+     in_index     = 0;
+     in_max       = 0;
+     in_buffer    = "";
+     lex_index    = 0;
+
+     out_max      = 0;
+     out_expand   = true;
+     out_buffer   = String.create buf_size;
+
+     write_pid    = 0;
+     write_index  = 0;
+     write_max    = 0;
+     write_buffer = "";
+
+     read_fun     = null_reader;
+     write_fun    = null_writer
+   }
+
+let create_loc_string file line char =
+   create_loc_string_aux (Lm_symbol.add file) line char
+
+let create_string () =
+   create_loc_string_aux string_sym 1 0
+
 (************************************************************************
  * Line envding translation.
  *)
@@ -463,6 +551,31 @@ let reset_output_buffer info =
    info.write_index  <- 0;
    info.write_max    <- 0
 
+(************************************************************************
+ * Output string buffers.
+ *)
+
+(*
+ * For string buffers, expand the output instead of
+ * flushing.
+ *)
+let expand_output info =
+   let { out_buffer = buffer;
+         out_max    = max
+       } = info
+   in
+      if max = String.length buffer then
+         let buffer2 = String.create (max * 2) in
+            String.blit buffer 0 buffer2 0 max;
+            info.out_buffer <- buffer2
+
+let to_string info =
+   let { out_buffer = buffer;
+         out_max = max
+       } = info
+   in
+      String.sub buffer 0 max
+
 (************************************************************************
  * Flushing and filling.
  *)
@@ -515,7 +628,7 @@ let flush_output_once info =
 let flush_aux info =
    setup_write_buffer info;
    let { write_buffer = buf;
-         write_fun  = writer
+         write_fun    = writer
        } = info
    in
 
@@ -538,7 +651,10 @@ let flush_output info =
    let pid = info.write_pid in
       if pid <> 0 then
          Lm_thread_pool.waitpid pid;
-      flush_aux info
+      if info.out_expand then
+         expand_output info
+      else
+         flush_aux info
 
 (*
  * Start an output thread trying to write the data to the pipe.
@@ -585,7 +701,7 @@ let rec output_char info c =
        } = info
    in
       flush_input info;
-      if max = buf_size then
+      if max = String.length buffer then
          begin
             flush_output info;
             output_char info c
@@ -607,7 +723,7 @@ let rec output_buffer info buf off len =
          out_buffer = buffer
        } = info
    in
-   let avail = buf_size - max in
+   let avail = String.length buffer - max in
       flush_input info;
       if len <> 0 then
          if avail = 0 then
@@ -862,7 +978,7 @@ let rec find_write_empty l wfd =
    match wfd with
       fd :: wfd ->
          let l =
-            if fd.out_max = 0 then
+            if fd.out_max = 0 || fd.out_expand then
                fd :: l
             else
                l
@@ -973,7 +1089,7 @@ let rec select_poll rfd_pipes rfd_sockets wfd_pipes wfd_sockets efd_sockets expi
 let select rfd wfd efd timeout =
    let rfd_files, rfd_pipes, rfd_sockets = classify [] [] [] rfd in
    let wfd_files, wfd_pipes, wfd_sockets = classify [] [] [] wfd in
-   let efd_files, efd_pipes, efd_sockets = classify [] [] [] efd in
+   let efd_files, _efd_pipes, efd_sockets = classify [] [] [] efd in
       if rfd_files <> [] || wfd_files <> [] || efd_files <> [] then
          rfd_files, wfd_files, efd_files
       else
@@ -1181,12 +1297,9 @@ struct
          channel.in_index <- max
 end
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
