@@ -4,71 +4,40 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2005 Mojave Group, Caltech
+ * Copyright (C) 2005-2007 Mojave Group, Caltech and HRL Laboratories, LLC
  *
- * This program is free software; you can redistribute it and/or
- * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * This library is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation,
+ * version 2.1 of the License.
  *
- * This program is distributed in the hope that it will be useful,
+ * This library is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * You should have received a copy of the GNU Lesser General Public
+ * License along with this library; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Additional permission is given to link this library with the
+ * OpenSSL project's "OpenSSL" library, and with the OCaml runtime,
+ * and you may distribute the linked executables.  See the file
+ * LICENSE.libmojave for more details.
+ *
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{anogin@hrl.com} @email{nogin@metaprl.org}
  * @end[license]
  *)
 open Lm_printf
 open Lm_thread_core
 
+open Lm_hash_sig
+
 (************************************************************************
  * A generic hash module to make comparisons faster.
  * This version uses a state for hash-consing.
  *)
-
-(*
- * The client needs to provide these functions.
- *)
-module type HashArgSig =
-sig
-   type t
-
-   (* For debugging *)
-   val debug : string
-
-   (* The client needs to provide hash and comparison functions *)
-   val hash : t -> int
-   val compare : t -> t -> int
-end
-
-(*
- * This is what we get.
- *)
-module type HashSig =
-sig
-   type elt
-   type t
-
-   (* Creation *)
-   val create : elt -> t
-   val get : t -> elt
-
-   (* Hash code *)
-   val hash : t -> int
-
-   (* Comparison *)
-   val compare : t -> t -> int
-end
-
-(*
- * Make a hash item.
- *)
 module MakeHash (Arg : HashArgSig)
 : HashSig with type elt = Arg.t =
 struct
@@ -94,43 +63,11 @@ struct
          -1
       else
          1
-end;;
+end
 
 (************************************************************************
  * Table-based hashing.
  *)
-module type HashConsSig =
-sig
-   type hash
-   type state
-   type elt
-   type t
-
-   (* States *)
-   val create_state : unit -> state
-   val length : state -> int
-
-   (* Normal creation *)
-   val icreate : state -> hash -> t
-   val create : state -> elt -> t
-   val get : state -> t -> elt
-
-   (* Hash code *)
-   val hash : t -> int
-
-   (* Comparison *)
-   val compare : t -> t -> int
-
-   (* Map over an array of hash codes *)
-   val map_array : (t -> elt -> 'a) -> state -> 'a array
-
-   (* Fold over all of the items *)
-   val fold : ('a -> t -> 'a) -> 'a -> state -> 'a
-end
-
-(*
- * This provides hash-consing.
- *)
 module MakeHashCons (Arg : HashArgSig)
 : HashConsSig
   with type elt = Arg.t
@@ -206,7 +143,7 @@ struct
             fold (succ i) (f x i)
       in
          fold 0 x
-end;;
+end
 
 (************************************************************************
  * Marshalable version.
@@ -217,22 +154,6 @@ end;;
  * must be rehashed.
  *)
 
-(*
- * The client needs to provide these functions.
- *)
-module type HashMarshalArgSig =
-sig
-   type t
-
-   (* For debugging *)
-   val debug : string
-
-   (* The client needs to provide hash and comparison functions *)
-   val hash : t -> int
-   val compare : t -> t -> int
-   val reintern : t -> t
-end
-
 (* %%MAGICBEGIN%% *)
 type 'a hash_marshal_item =
    { mutable item_ref : unit ref;
@@ -241,31 +162,6 @@ type 'a hash_marshal_item =
    }
 (* %%MAGICEND%% *)
 
-(*
- * This is what we get.
- *)
-module type HashMarshalSig =
-sig
-   type elt
-   type t = elt hash_marshal_item
-
-   (* Creation *)
-   val create   : elt -> t
-
-   (* The intern function fails with Not_found if the node does not already exist *)
-   val intern   : elt -> t
-
-   (* Destructors *)
-   val get      : t -> elt
-   val hash     : t -> int
-
-   (* Comparison *)
-   val compare  : t -> t -> int
-
-   (* Rehash the value *)
-   val reintern : t -> t
-end;;
-
 (*
  * The reference in the current process.
  *)
@@ -290,17 +186,21 @@ let pp_print_stat buf stat =
          hash_collisions = collisions
        } = stat
    in
-      fprintf buf "@ @[<hv 3>%s: reintern = %d, compare = %d, collisions = %d@]" (**)
+      fprintf buf "@[<hv 3>%s: reintern = %d, compare = %d, collisions = %d@]@\n" (**)
          debug reintern compare collisions
 
 let pp_print_hash_stats buf =
    List.iter (pp_print_stat buf) !hash_stats
 
+(*
+ * let () =
+ *    at_exit (fun () -> pp_print_hash_stats stderr)
+ *)
+
 (*
  * Make a hash item.
  *)
-module MakeHashMarshal (Arg : HashMarshalArgSig)
-: HashMarshalSig with type elt = Arg.t =
+module MakeHashMarshal (Arg : HashMarshalArgSig) =
 struct
    type elt = Arg.t
    type t = elt hash_marshal_item
@@ -469,10 +369,106 @@ struct
                   stats.hash_collisions <- succ stats.hash_collisions;
                   let cmp = Arg.compare elt1 elt2 in
                      if cmp = 0 then
-                        eprintf "Hash is broken@.";
+                        invalid_arg "Lm_hash is broken@.";
                      cmp
                end
-end;;
+
+   let equal item1 item2 =
+      (item1 == item2) || (item1.item_hash = item2.item_hash && get item1 == get item2)
+end
+
+(*
+ * A version with two equalities.
+ * The fine equality is used for cons-hashing, but the coarse
+ * version is used for external comparisons.  The fine equality
+ * must be a refinement of the coarse equality.
+ *)
+(* %%MAGICBEGIN%% *)
+type 'a hash_marshal_eq_item = ('a * 'a hash_marshal_item) hash_marshal_item
+(* %%MAGICEND%% *)
+
+module MakeHashMarshalEq (Arg : HashMarshalEqArgSig) =
+struct
+   type elt = Arg.t
+   type t = elt hash_marshal_eq_item
+
+   module CoarseArg =
+   struct
+      type t     = Arg.t
+      let debug  = Arg.debug ^ ":coarse"
+
+      let hash     = Arg.coarse_hash
+      let compare  = Arg.coarse_compare
+      let reintern = Arg.reintern
+   end;;
+
+   module Coarse = MakeHashMarshal (CoarseArg);;
+
+   (*
+    * We fold the Coarse item into the fine
+    * item only so we don't have to create three
+    * modules (the final one being a pair of fine
+    * and coarse).
+    *)
+   module FineArg =
+   struct
+      type t     = Arg.t * Coarse.t
+      let debug  = Arg.debug ^ ":fine"
+
+      (*
+       * We're assuming that the fine hash is a
+       * refinement of the coarse one.
+       *)
+      let hash (fine, _) =
+         Arg.fine_hash fine
+
+      let compare (fine1, _) (fine2, _) =
+         Arg.fine_compare fine1 fine2
+
+      let reintern ((fine, coarse) as item) =
+         let fine' = Arg.reintern fine in
+         let coarse' = Coarse.reintern coarse in
+            if fine' == fine && coarse' == coarse then
+               item
+            else
+               fine', coarse'
+   end;;
+
+   module Fine = MakeHashMarshal (FineArg);;
+
+   let create x =
+      Fine.create (x, Coarse.create x)
+
+   let intern x =
+      Fine.intern (x, Coarse.intern x)
+
+   let get info =
+      fst (Fine.get info)
+
+   (*
+    * The default functions are the coarse versions.
+    *)
+   let get_coarse info =
+      snd (Fine.get info)
+
+   let hash info =
+      Coarse.hash (get_coarse info)
+
+   let compare item1 item2 =
+      Coarse.compare (get_coarse item1) (get_coarse item2)
+
+   let equal item1 item2 =
+      Coarse.equal (get_coarse item1) (get_coarse item2)
+
+   (*
+    * Also export the fine versions.
+    *)
+   let fine_hash = Fine.hash
+   let fine_compare = Fine.compare
+   let fine_equal = Fine.equal
+
+   let reintern = Fine.reintern
+end
 
 (************************************************************************
  * Better-than usual hash functions.
@@ -485,7 +481,7 @@ end;;
  *)
 
 (* %%MAGICBEGIN%% *)
-let hash_length = 6217
+let hash_length = 6229 (* Must be a prime *)
 let digest_length = 16
 let hash_data =
    [|0x04a018c6; 0x5ba7b0f2; 0x04dcf08b; 0x1e5a22cc; 0x2523b9ea; 0x4b92b691; 0x0bcaf015; 0x0a5d5109;
@@ -1269,23 +1265,15 @@ let hash_data =
      0x1a1f3371; 0x61cc8fb3; 0x3711c483; 0x28a61268; 0x2ed93c6e; 0x2e42ed8c; 0x1a387531; 0x6b1bd449;
      0x301efe32; 0x5dcf70d0; 0x3dd03c00; 0x1153313d; 0x129c2b48; 0x3b9d34b9; 0x2ba65435; 0x6e0426bb;
      0x1955c17b; 0x4f0e1b76; 0x028d41c9; 0x02d475d3; 0x2ef7c30f; 0x2ea73087; 0x357c92a5; 0x4141569c;
-     0x27f6f470|]
+     |]
 (* %%MAGICEND%% *)
 
+let () =
+   assert(Array.length hash_data >= hash_length + digest_length - 1)
+
 (************************************************
  * Integer hashes.
  *)
-module type HashCodeSig =
-sig
-   type t
-
-   val create     : unit -> t
-   val add_int    : t -> int -> unit
-   val add_float  : t -> float -> unit
-   val add_string : t -> string -> unit
-   val code       : t -> int
-end;;
-
 (* %%MAGICBEGIN%% *)
 module HashCode : HashCodeSig =
 struct
@@ -1305,37 +1293,50 @@ struct
    (*
     * Add an integer.
     *)
-   let add_bits buf i =
+   let add_truncated_bits buf i =
       let { hash_digest = digest;
             hash_code   = code
           } = buf
       in
-      let code = ((code + i + 1) land 0x3fffffff) mod hash_length in
-         buf.hash_digest <- (digest lsl 3) lxor (digest lsr 1) lxor (Array.unsafe_get hash_data code);
+      let code = (code + i + 1) mod hash_length in
+         buf.hash_digest <- digest lxor (Array.unsafe_get hash_data code);
          buf.hash_code <- code
 
+   let add_bits buf i =
+      add_truncated_bits buf (i land 0x7ff)
+
    (*
     * Add the characters in a string.
     *)
    let add_string buf s =
       for i = 0 to pred (String.length s) do
-         add_bits buf (Char.code (String.unsafe_get s i))
+         add_truncated_bits buf (Char.code (String.unsafe_get s i))
       done
 
    (*
     * Numbers.
     *)
    let add_int buf i =
-      add_bits buf (i land 0xff);
-      add_bits buf ((i lsr 8) land 0xff);
-      add_bits buf ((i lsr 16) land 0xff);
-      add_bits buf ((i lsr 24) land 0xff)
+      add_bits buf i;
+      add_bits buf (i lsr 11);
+      add_bits buf (i lsr 22)
+
+   let add_int32 buf i =
+      add_bits buf (Int32.to_int (Int32.shift_right_logical i 16));
+      add_bits buf (Int32.to_int i)
+
+   let add_int64 buf i =
+      add_int buf (Int64.to_int (Int64.shift_right_logical i 48));
+      add_int buf (Int64.to_int (Int64.shift_right_logical i 24));
+      add_int buf (Int64.to_int i)
+
+   let add_nativeint buf i =
+      add_int buf (Nativeint.to_int (Nativeint.shift_right_logical i 48));
+      add_int buf (Nativeint.to_int (Nativeint.shift_right_logical i 24));
+      add_int buf (Nativeint.to_int i)
 
    let add_float buf x =
-      let i = Int64.bits_of_float x in
-         add_int buf (Int64.to_int (Int64.shift_right i 6));
-         add_int buf (Int64.to_int (Int64.shift_right i 3));
-         add_int buf (Int64.to_int i)
+      add_int64 buf (Int64.bits_of_float x)
 
    (*
     * Extract the digest.
@@ -1348,25 +1349,11 @@ end
 (************************************************
  * Digest-based hashes.
  *)
-module type HashDigestSig =
-sig
-   type t
-
-   val create     : unit -> t
-   val add_bits   : t -> int -> unit
-   val add_bool   : t -> bool -> unit
-   val add_int    : t -> int -> unit
-   val add_float  : t -> float -> unit
-   val add_string : t -> string -> unit
-   val digest     : t -> string
-end;;
-
 (* %%MAGICBEGIN%% *)
 module HashDigest : HashDigestSig =
 struct
    type t =
       { hash_digest         : int array;
-        mutable hash_length : int;              (* For stats *)
         mutable hash_code   : int
       }
 
@@ -1375,51 +1362,73 @@ struct
     *)
    let create () =
       { hash_digest = Array.create digest_length 0;
-        hash_length = 0;
         hash_code   = 0
       }
 
    (*
     * Add an integer.
     *)
-   let add_bits buf i =
+   let add_truncated_bits buf i =
       let { hash_digest = digest;
-            hash_length = length;
             hash_code   = code
           } = buf
       in
-      let code = ((code + digest_length + i) land 0x3fffffff) mod hash_length in
+      let code = (code + digest_length + i) mod hash_length in
          for i = 0 to digest_length - 1 do
-            let v = digest.(i) in
-               digest.(i) <- (v lsl 3) lxor (v lsr 1) lxor (Array.unsafe_get hash_data (code + i))
+            digest.(i) <- digest.(i) lxor (Array.unsafe_get hash_data (code + i))
          done;
-         buf.hash_code <- code;
-         buf.hash_length <- succ length
+         buf.hash_code <- code
+
+   let add_bits buf i =
+      add_truncated_bits buf (i land 0x7ff)
 
    (*
     * Add the characters in a string.
     *)
+   let add_char buf c =
+      add_truncated_bits buf (Char.code c)
+
    let add_string buf s =
       for i = 0 to pred (String.length s) do
-         add_bits buf (Char.code (String.unsafe_get s i))
+         add_char buf (String.unsafe_get s i)
+      done
+
+   let add_substring buf s off len =
+      if off < 0 || len < 0 || off + len > String.length s then
+         raise (Invalid_argument "Lm_hash.add_substring");
+      for i = off to pred (off + len) do
+         add_char buf (String.unsafe_get s i)
       done
 
    (*
     * Numbers.
     *)
    let add_bool buf b =
-      add_bits buf (if b then 1 else 0)
+      add_truncated_bits buf (if b then 1 else 0)
 
    let add_int buf i =
-      add_bits buf (i land 0xff);
-      add_bits buf ((i lsr 8) land 0xff);
-      add_bits buf ((i lsr 16) land 0xff);
-      add_bits buf ((i lsr 24) land 0xff)
+      add_bits buf i;
+      add_bits buf (i lsr 11);
+      add_bits buf (i lsr 22)
+
+   let add_int32 buf i =
+      add_bits buf (Int32.to_int (Int32.shift_right_logical i 16));
+      add_bits buf (Int32.to_int i)
+
+   let add_int64 buf i =
+      add_int buf (Int64.to_int (Int64.shift_right_logical i 48));
+      add_int buf (Int64.to_int (Int64.shift_right_logical i 24));
+      add_int buf (Int64.to_int i)
+
+   let add_nativeint buf i =
+      add_int buf (Nativeint.to_int (Nativeint.shift_right_logical i 48));
+      add_int buf (Nativeint.to_int (Nativeint.shift_right_logical i 24));
+      add_int buf (Nativeint.to_int i)
 
    let add_float buf x =
       let i = Int64.bits_of_float x in
-         add_int buf (Int64.to_int (Int64.shift_right i 6));
-         add_int buf (Int64.to_int (Int64.shift_right i 3));
+         add_int buf (Int64.to_int (Int64.shift_right i 48));
+         add_int buf (Int64.to_int (Int64.shift_right i 24));
          add_int buf (Int64.to_int i)
 
    (*
@@ -1441,8 +1450,8 @@ end;;
 
 (*
  * The default function for combinding hash values.
- *  * XXX: JYH: we should try using a smarter hash function.
- *   *)
+ * XXX: JYH: we should try using a smarter hash function.
+ *)
 let hash_combine i1 i2 =
    (i1 lsl 2) lxor (i1 lsr 2) lxor i2
 
@@ -1474,12 +1483,9 @@ let rec compare_int_list (l1 : int list) (l2 : int list) =
     | [], [] ->
          0
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
