@@ -4,24 +4,28 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2003 Jason Hickey, Caltech
+ * Copyright (C) 2003-2006 Mojave Group, Caltech
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
- * Author: Jason Hickey
- * @email{jyh@cs.caltech.edu}
+ * Author: Jason Hickey @email{jyh@cs.caltech.edu}
+ * Modified By: Aleksey Nogin @email{nogin@cs.caltech.edu}
  * @end[license]
  *)
 open Lm_printf
@@ -36,11 +40,20 @@ open Omake_rule
 open Omake_state
 open Omake_symbol
 open Omake_exec_type
-open Omake_options_type
+open Omake_options
 
-module Pos = MakePos (struct let name = "Omake_main" end)
+module Pos = MakePos (struct let name = "Omake_main" end);;
 open Pos
 
+(*
+ * Print the hash stats?
+ *)
+let debug_hash = ref false
+
+let print_hash_stats () =
+   if !debug_hash then
+      eprintf "@[<v 3>Hash statistics:@ %t@]@." Lm_hash.pp_print_hash_stats
+
 (*
  * List of targets to build.
  *)
@@ -93,22 +106,29 @@ let spec =
    Lm_arg.StrictOptions, (**)
       ["Make flags", (**)
           options_spec;
+       "Output flags", (**)
+          output_spec;
        "Cache management", (**)
-          ["--force-dotomake", Lm_arg.Set Omake_state.always_use_dotomake, (**)
+          ["--save-interval", Lm_arg.Float (fun f -> Omake_build.save_interval := f), (**)
+              (sprintf "Save the build DB (\".omakedb\") every x seconds (0 disables, default: %F)" (**)
+                  Omake_magic.default_save_interval);
+           "--force-dotomake", Lm_arg.Set Omake_state.always_use_dotomake, (**)
               "Always use $HOME/.omake for .omc cache files";
            "--dotomake", Lm_arg.String (fun s -> Omake_state.set_omake_dir s), (**)
               "Use the specified directory in place of $HOME/.omake"];
        "Helper flags", (**)
           ["--install", Lm_arg.Set install_flag, (**)
               "Install an OMake project into the current directory";
-           "--install-all", Lm_arg.Unit (fun () ->
-                 install_flag := true;
-                 install_subdirs := true), (**)
+           "--install-all", Lm_arg.SetFold (fun opts b ->
+                 install_flag := b;
+                 install_subdirs := b;
+                 opts), (**)
               "Install an OMake project into the current directory and all subdirectories";
-           "--install-force", Lm_arg.Unit (fun () ->
-                 install_flag := true;
-                 install_force := true), (**)
-              "Force overwriting of files during installation";
+           "--install-force", Lm_arg.SetFold (fun opts b ->
+                 install_flag := b;
+                 install_force := b;
+                 opts), (**)
+              "Force overwriting of files during installation; implies --install";
            "--version", Lm_arg.Unit (fun () ->
                  printf "%s\n\nDefault library directory : %s@." Omake_magic.version_message Omake_magic.lib_dir;
                  if Omake_state.lib_dir_reason <> "" then
@@ -119,7 +139,7 @@ let spec =
        "Shell flags", (**)
           ["--shell", Lm_arg.Set shell_flag, (**)
               "Run the OMake shell: osh";
-           "-i", Lm_arg.Unit (fun () -> Omake_readline.set_interactive true), (**)
+           "-i", Lm_arg.SetFold (fun opts b -> Omake_readline.set_interactive b; opts), (**)
               "Treat the session as interactive";
            "-c", Lm_arg.String (fun s -> command_string := Some s), (**)
               "Evaluate the commands from the string"];
@@ -172,11 +192,13 @@ let spec =
               "Debug the parser generator";
            "-debug-parsing", Lm_arg.Set Omake_builtin_io_fun.debug_parsing, (**)
               "Debug OMake parsing operations";
-           "-debug-notify", Lm_arg.Set Omake_notify.debug_notify, (**)
+           "-debug-notify", Lm_arg.Set Lm_notify.debug_notify, (**)
               "Debug the FAM (-p filesystem watch) operations";
            "-debug-db", Lm_arg.Set Omake_env.debug_db, (**)
               "Debug the file database";
-           "-allow-exceptions", Lm_arg.UnitFold (fun options -> { options with opt_allow_exceptions = true }),
+           "-debug-hash", Lm_arg.Set debug_hash, (**)
+              "Show Lm_hash statistics";
+           "-allow-exceptions", Lm_arg.SetFold set_allow_exceptions_opt, (**)
               "Do not catch top-level exceptions (for use with OCAMLRUNPARAM=b)"];
        "Internal flags", (**)
           ["-server", Lm_arg.String (fun s -> server_flag := Some s), (**)
@@ -209,20 +231,22 @@ let chroot () =
    let cwd =
       try
          Unix.getcwd ()
-      with Unix.Unix_error _ ->
-         eprintf "*** omake: fatal error: current directory does not exist@.";
-         exit 1
+      with
+         Unix.Unix_error _ ->
+            eprintf "*** omake: fatal error: current directory does not exist@.";
+            exit 1
    in
    let len = String.length cwd in
    let start = Lm_filename_util.drive_skip cwd in
    let rec search i =
       if i < start then
-         raise (OmakeFatal ("can not find " ^ makeroot_name))
+         raise (OmakeFatal ("can not find " ^ makeroot_name ^ " or " ^ makeroot_short_name))
       else if cwd.[i] = '/' || cwd.[i] = '\\' then
          (* Maybe file is in this directory *)
          let dir = String.sub cwd 0 i in
-         let filename = Filename.concat dir makeroot_name in
-            if Sys.file_exists filename then
+            if Sys.file_exists (Filename.concat dir makeroot_name)
+               || Sys.file_exists (Filename.concat dir makeroot_short_name)
+            then
                let rest = String.sub cwd (succ i) (len - i - 1) in
                   dir, rest
             else
@@ -231,11 +255,11 @@ let chroot () =
          search (pred i)
    in
    let dir, rest =
-      let filename = Filename.concat cwd makeroot_name in
-         if Sys.file_exists filename then
-            cwd, "."
-         else
-            search (pred len)
+      if Sys.file_exists (Filename.concat cwd makeroot_name)
+         || Sys.file_exists (Filename.concat cwd makeroot_short_name) then
+         cwd, "."
+      else
+         search (pred len)
    in
       if rest <> "." then
          eprintf "*** omake: changing directory to %s@." dir;
@@ -254,7 +278,7 @@ let main options =
    in
    let path = chroot () in
    let path =
-      if options.opt_cd_root then
+      if opt_cd_root options then
          "."
       else
          path
@@ -264,7 +288,8 @@ let main options =
          [] -> [".DEFAULT"]
        | l -> List.rev l
    in
-      Omake_build.build options path targets
+      Omake_build.build options path targets;
+      print_hash_stats ()
 
 (*
  * Main for remote execution.
@@ -328,17 +353,14 @@ let _ =
                   Omake_install.install_subdirs !install_force
                else
                   Omake_install.install_current !install_force
-            else if options.opt_allow_exceptions then
+            else if opt_allow_exceptions options then
                main options
             else
                Omake_exn_print.catch main options
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
