@@ -4,12 +4,12 @@
  * ----------------------------------------------------------------
  *
  * @begin[license]
- * Copyright (C) 2004-2006 Mojave Group, Caltech
+ * Copyright (C) 2004-2007 Mojave Group, Caltech and HRL Laboratories, LLC
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
+ * as published by the Free Software Foundation; version 2
+ * of the License.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
@@ -18,15 +18,20 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ *
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey @email{jyh@cs.caltech.edu}
- * Modified by: Aleksey Nogin @email{nogin@cs.caltech.edu}
+ * Modified by: Aleksey Nogin @email{nogin@cs.caltech.edu}, @email{anogin@hrl.com}
  * @end[license]
  *)
 open Lm_printf
 open Lm_symbol
 open Lm_location
+open Lm_string_set
 
 open Omake_ir
 open Omake_env
@@ -70,7 +75,7 @@ type job_status =
 
 type subjob_cond =
    { cond_op       : pipe_op;
-     cond_pipe     : pipe;
+     cond_pipe     : string_pipe;
      cond_stdin    : Unix.file_descr;
      cond_stdout   : Unix.file_descr;
      cond_stderr   : Unix.file_descr
@@ -89,7 +94,7 @@ and subjob_exp =
  *)
 type job =
    { job_id              : int;
-     job_pipe            : pipe option;
+     job_pipe            : string_pipe option;
      mutable job_pgrp    : pgrp;
      mutable job_state   : job_state
    }
@@ -128,116 +133,12 @@ let pp_print_job_state buf state =
       pp_print_string buf s
 
 (*
- * Print an application.
+ * Job may be a pipe.
  *)
-let rec pp_print_args buf args =
-   match args with
-      [arg] ->
-         pp_print_string buf arg
-    | arg :: args ->
-         pp_print_string buf arg;
-         pp_print_string buf ", ";
-         pp_print_args buf args
-    | [] ->
-         ()
-
-let pp_print_apply buf apply =
-   let { apply_name = f;
-         apply_args = args
-       } = apply
-   in
-      fprintf buf "@[<hv 3>apply %a(%a)@]" pp_print_symbol f pp_print_args args
-
-(*
- * Print a command.
- *)
-let pp_print_command buf command =
-   let { cmd_exe = exe;
-         cmd_env = env;
-         cmd_argv = argv;
-         cmd_stdin = stdin;
-         cmd_stdout = stdout;
-         cmd_stderr = stderr
-       } = command
-   in
-      fprintf buf "@[<v 3>command@ exe = %s" exe;
-      (match stdin with
-          Some stdin ->
-             fprintf buf "@ stdin = %s" stdin
-        | None ->
-             ());
-      (match stdout with
-          Some stdout ->
-             fprintf buf "@ stdout = %s" stdout
-        | None ->
-             ());
-      if stderr then
-         fprintf buf "@ stderr = stdout";
-      List.iter (fun (v, x) ->
-            fprintf buf "@ %a=%s" pp_print_symbol v x) env;
-      List.iter (fun v ->
-            fprintf buf "@ %s" v) argv;
-      fprintf buf "@]"
-
-(*
- * Print a pipe.
- *)
-let pp_print_pipe_op buf op =
-   let s =
-      match op with
-         PipeAnd -> "&&"
-       | PipeOr -> "||"
-       | PipeSequence -> ";"
-   in
-      pp_print_string buf s
-
-let rec pp_print_pipe buf pipe =
-   match pipe with
-      PipeApply (_, apply) ->
-         pp_print_apply buf apply
-    | PipeCommand (_, command) ->
-         pp_print_command buf command
-    | PipeCond (_, op, pipe1, pipe2) ->
-         fprintf buf "%a %a %a" (**)
-            pp_print_pipe pipe1
-            pp_print_pipe_op op
-            pp_print_pipe pipe2
-    | PipeCompose (_, divert_stderr, pipe1, pipe2) ->
-         fprintf buf "%a %s %a" (**)
-            pp_print_pipe pipe1
-            (if divert_stderr then "|&" else "|")
-            pp_print_pipe pipe2
-    | PipeGroup (_, group) ->
-         pp_print_group buf group
-    | PipeBackground (_, pipe) ->
-         pp_print_pipe buf pipe
-
-and pp_print_group buf group =
-   let { group_stdin = stdin;
-         group_stdout = stdout;
-         group_stderr = stderr;
-         group_pipe = pipe
-       } = group
-   in
-      fprintf buf "@[<v 3>group@ %a" pp_print_pipe pipe;
-      (match stdin with
-          Some stdin ->
-             fprintf buf "@ stdin = %s" stdin
-        | None ->
-             ());
-      (match stdout with
-          Some stdout ->
-             fprintf buf "@ stdout = %s" stdout
-        | None ->
-             ());
-      if stderr then
-         fprintf buf "@ stderr = stdout";
-      fprintf buf "@]"
-
 let pp_print_pipe_option buf opt =
    match opt with
       Some pipe ->
-         pp_print_pipe buf pipe
+         pp_print_string_pipe buf pipe
     | None ->
          pp_print_string buf "<thread>"
 
@@ -270,7 +171,7 @@ let rec pp_print_exp buf e =
             fprintf buf "@[<hv 1>(%a)@ %a@ %a@]" (**)
                pp_print_exp e
                pp_print_pipe_op op
-               pp_print_pipe pipe
+               pp_print_string_pipe pipe
     | SubjobFinished (code, _) ->
          fprintf buf "[Finished: %a]" pp_print_status code
 
@@ -449,16 +350,25 @@ let create_top_thread venv f stdin stdout stderr =
 (*
  * Create the diversion channels.
  *)
+let string_of_redirect chan =
+   match chan with
+      RedirectNode node ->
+         Some (Node.fullname node)
+    | RedirectArg s ->
+         Some s
+    | RedirectNone ->
+         None
+
 let create_channels stdin stdin_file append stdout stdout_file stderr_divert stderr =
    let stdin, close_stdin =
-      match stdin_file with
+      match string_of_redirect stdin_file with
          Some file ->
             Lm_unix_util.openfile file [Unix.O_RDONLY; Unix.O_NOCTTY] 0, true
        | None ->
             stdin, false
    in
    let stdout, close_stdout =
-      match stdout_file with
+      match string_of_redirect stdout_file with
          Some file ->
             let flags = [Unix.O_WRONLY; Unix.O_CREAT; Unix.O_NOCTTY] in
             let flags =
@@ -491,9 +401,11 @@ let create_channels stdin stdin_file append stdout stdout_file stderr_divert std
  * Application at the toplevel.
  * Don't create a thread.
  *)
-let restore_vars = [ stdin_sym; stdout_sym; stderr_sym ]
+let restore_vars = [stdin_sym; stdout_sym; stderr_sym]
+
 let create_apply_top venv stdin stdout stderr apply =
    let { apply_loc = loc;
+         apply_env = env;
          apply_fun = f;
          apply_args = args;
          apply_stdin = stdin_file;
@@ -514,20 +426,20 @@ let create_apply_top venv stdin stdout stderr apply =
       (* The function will close its files on its own *)
       try
          if !debug_shell then
-            eprintf "create_apply_top: duplicating channels@.";
+            eprintf "create_apply_top: %i duplicating channels@." (Unix.getpid ());
          let stdin  = Unix.dup stdin in
          let stdout = Unix.dup stdout in
          let stderr = Unix.dup stderr in
-         let code, v = f venv stdin stdout stderr args in
+         let code, v = f venv stdin stdout stderr env args in
          let v = unexport venv v restore_vars in
             if !debug_shell then
-               eprintf "create_apply_top: done@.";
+               eprintf "create_apply_top %i: done@." (Unix.getpid ());
             cleanup ();
             code, v
       with
          exn ->
             if !debug_shell then
-               eprintf "create_apply_top: error@.";
+               eprintf "create_apply_top %i: error: %a@." (Unix.getpid ()) Omake_exn_print.pp_print_exn exn;
             cleanup ();
             raise exn
 
@@ -538,6 +450,7 @@ let create_apply venv pgrp bg stdin stdout stderr apply =
    if !debug_shell then
       eprintf "create_apply@.";
    let { apply_loc = loc;
+         apply_env = env;
          apply_fun = f;
          apply_args = args;
          apply_stdin = stdin_file;
@@ -552,7 +465,7 @@ let create_apply venv pgrp bg stdin stdout stderr apply =
 
    (* The actual function call *)
    let apply_fun stdin stdout stderr pgrp =
-      fst (f venv stdin stdout stderr args)
+      fst (f venv stdin stdout stderr env args)
    in
    let thread_info =
       { create_thread_stdin = stdin;
@@ -581,35 +494,40 @@ let create_apply venv pgrp bg stdin stdout stderr apply =
 (*
  * Resolve the absolute name of the executable.
  *)
-let find_executable venv pos loc exe =
+let find_executable_string venv pos loc exe =
    let pos = string_pos "find_executable" pos in
    let cache = venv_cache venv in
-   let node =
       if not (Filename.is_relative exe) || Lm_string_util.contains_any exe Lm_filename_util.separators then
-         let node = venv_intern venv PhonyProhibited exe in
-            if Omake_cache.exists cache node then
-               node
-            else if (Sys.os_type = "Win32" || Sys.os_type = "Cygwin")
-               && not (Filename.check_suffix exe ".exe" || Filename.check_suffix exe ".bat")
-            then
-               let node = venv_intern venv PhonyProhibited (exe ^ ".exe") in
+         let rec resolve_exe = function
+            suff :: suffixes ->
+               let node = venv_intern venv PhonyProhibited (exe ^ suff) in
                   if Omake_cache.exists cache node then
                      node
                   else
-                     let node = venv_intern venv PhonyProhibited (exe ^ ".bat") in
-                        if Omake_cache.exists cache node then
-                           node
-                        else
-                           raise (OmakeException (loc_pos loc pos, StringStringError ("command not found", exe)))
-            else
+                     resolve_exe suffixes
+          | [] ->
                raise (OmakeException (loc_pos loc pos, StringStringError ("command not found", exe)))
+         in
+            resolve_exe Omake_cache.exe_suffixes
       else
          let path = venv_find_var venv ScopeGlobal pos loc path_sym in
          let path = Omake_eval.path_of_values venv pos (values_of_value venv pos path) "." in
          let path = Omake_cache.ls_exe_path cache path in
             try Omake_cache.exe_find cache path exe with
                Not_found ->
-                  raise (OmakeException (loc_pos loc pos, StringStringError ("command not found", exe)))
+                  raise (OmakeException (loc_pos loc pos, StringStringError ("command not found in PATH", exe)))
+
+let find_executable venv pos loc exe =
+   let node =
+      match exe with
+         ExeQuote exe
+       | ExeString exe ->
+            find_executable_string venv pos loc exe
+       | ExeNode node ->
+            if Omake_cache.exe_suffixes = ["."] then
+               node
+            else
+               find_executable_string venv pos loc (Node.absname node)
    in
       Node.absname node
 
@@ -644,7 +562,7 @@ let create_command venv pgrp bg stdin stdout stderr command =
         create_process_env    = array_of_env current_env env;
         create_process_dir    = dir;
         create_process_exe    = exe;
-        create_process_argv   = Array.of_list argv;
+        create_process_argv   = Array.of_list (exe :: argv);
         create_process_background = bg
       }
    in
@@ -665,7 +583,23 @@ let create_command venv pgrp bg stdin stdout stderr command =
       with
          exn ->
             cleanup ();
-            raise exn
+            begin match exn with
+               Failure err ->
+                  let format_error buf =
+                     fprintf buf "@[<hv3>Spawning %s failed:@ %s@]" exe err
+                  in
+                     raise (OmakeException(pos, LazyError format_error))
+             | Unix.Unix_error(err, cmd, arg) ->
+                  let format_error buf =
+                     fprintf buf "@[<hv3>Spawning %s failed:@ @[<hv3>%s" exe cmd;
+                     if (arg <> "") then
+                        fprintf buf "@ %s" arg;
+                     fprintf buf ":@ %s@]@]" (Unix.error_message err)
+                  in
+                     raise (OmakeException(pos, LazyError format_error))
+             | _ ->
+                  raise exn
+            end
 
 (*
  * Create a conditional.
@@ -781,6 +715,8 @@ and create_group venv pgrp stdin stdout stderr group =
  * Create the pipe.
  *)
 and create_pipe_aux venv pgrp fork stdin stdout stderr pipe =
+   if !debug_shell then
+      eprintf "create_pipe_aux (fork: %b): %a@." fork pp_print_string_pipe pipe;
    match pipe with
       PipeApply (loc, apply) ->
          if fork then
@@ -848,19 +784,20 @@ and create_thread venv f stdin stdout stderr =
  * is exit.
  *)
 and wait_exp pgrp exp =
-   let exp = eval_exp pgrp exp 0 (JobExited 0) in
-      match exp with
-         SubjobFinished (JobExited code, _)
-       | SubjobFinished (JobSignaled code, _) ->
-            if !debug_shell then
-               eprintf "wait_exp: exiting %d@." code;
-            code
-       | exp ->
-            wait_exp2 pgrp exp
+   match eval_exp_top pgrp exp with
+      SubjobFinished (JobExited code, _)
+    | SubjobFinished (JobSignaled code, _) ->
+         if !debug_shell then
+            eprintf "wait_exp: %i exiting %d@." (Unix.getpid()) code;
+         code
+    | exp ->
+         wait_exp2 pgrp exp
 
 and wait_exp2 pgrp exp =
    (* Wait for a job to complete; ignore stopped processes *)
    let rec wait () =
+      if !debug_shell then
+         eprintf "wait_exp2: %i waiting for pgrp %i@." (Unix.getpid()) pgrp;
       let code =
          try Some (Omake_shell_sys.wait pgrp false false) with
             Unix.Unix_error (Unix.EINTR, _, _) ->
@@ -880,13 +817,13 @@ and wait_exp2 pgrp exp =
    let pid, code = wait () in
       (* Evaluate the expression *)
       if !debug_shell then
-         eprintf "wait_exp: handling event: pid=%d@." pid;
+         eprintf "wait_exp2: %i handling event: pid=%d@." (Unix.getpid()) pid;
       let exp = eval_exp pgrp exp pid code in
          match exp with
             SubjobFinished (JobExited code, _)
           | SubjobFinished (JobSignaled code, _) ->
                if !debug_shell then
-                  eprintf "wait_exp: exiting %d@." code;
+                  eprintf "wait_exp2: %i exiting %d@." (Unix.getpid()) code;
                code
           | exp ->
                wait_exp2 pgrp exp
@@ -894,7 +831,12 @@ and wait_exp2 pgrp exp =
 (*
  * Evaluate the expression.
  *)
+and eval_exp_top pgrp e =
+   eval_exp pgrp e 0 (JobExited 0)
+
 and eval_exp pgrp e pid code =
+   if !debug_shell then
+      eprintf "eval_exp in %i: pgrp=%i, pid=%i@." (Unix.getpid()) pgrp pid;
    let rec eval e =
       match e with
          SubjobProcess (pid', venv) ->
@@ -909,6 +851,7 @@ and eval_exp pgrp e pid code =
               | e1, e2 ->
                    SubjobPipe (e1, e2))
        | SubjobCond (e, cond) ->
+            if !debug_shell then eprintf "eval_exp in %i: evaluating SubjobCond@." (Unix.getpid());
             (match eval e with
                 SubjobFinished (code, venv) ->
                    let { cond_op     = op;
@@ -938,7 +881,7 @@ and eval_exp pgrp e pid code =
                             false
                    in
                       if cont then
-                         create_pipe_aux venv pgrp false stdin stdout stderr pipe
+                         eval_exp_top pgrp (create_pipe_aux venv pgrp false stdin stdout stderr pipe)
                       else
                          SubjobFinished (code, venv)
               | e ->
@@ -1001,9 +944,9 @@ let wait_pid venv job =
  *)
 let create_job venv pipe stdin stdout stderr =
    if !debug_shell then
-      eprintf "Creating pipe: %a@." pp_print_pipe pipe;
+      eprintf "Creating pipe: %a@." pp_print_string_pipe pipe;
 
-   (* Evaluate application eagerly *)
+   (* Evaluate applications eagerly *)
    match pipe with
       PipeApply (loc, apply) when stdout = Unix.stdout && stderr = Unix.stderr ->
          let _, value = create_apply_top venv stdin stdout stderr apply in
@@ -1013,18 +956,28 @@ let create_job venv pipe stdin stdout stderr =
          let bg   = is_background_pipe pipe in
          let pgrp = create_pipe_exn venv bg stdin stdout stderr pipe in
          let job  = new_job pgrp (Some pipe) in
-            if not bg then
-               begin
-                  if !debug_shell then
-                     eprintf "Running pgrp %d@." pgrp;
-                  Omake_shell_sys.set_tty_pgrp pgrp;
-                  ValOther (ValExitCode (int_of_code (wait_top venv job)))
-               end
-            else
-               begin
-                  job.job_state <- JobBackground;
-                  ValNone
-               end
+            if not bg then begin
+               if !debug_shell then
+                  eprintf "Running pgrp %d (my pid = %d)@." pgrp (Unix.getpid ());
+               (*
+                * On Mac OSX this call fails with EPERM.
+                * I believe this is because the sub-process
+                * sets the controlling terminal itself (see
+                * Omake_shell_sys_unix.create_process).
+                *
+                * This means that the sub-process takes over the terminal,
+                * and we can't set it anymore.
+                *
+                * This seems like a bogus explanation, because we have
+                * to get the terminal back on suspend...
+               Omake_shell_sys.set_tty_pgrp pgrp;
+                *)
+               ValOther (ValExitCode (int_of_code (wait_top venv job)))
+            end
+            else begin
+               job.job_state <- JobBackground;
+               ValNone
+            end
 
 (*
  * This is a variation: create the process and return the pid.
@@ -1032,7 +985,7 @@ let create_job venv pipe stdin stdout stderr =
  *)
 let create_process venv pipe stdin stdout stderr =
    if !debug_shell then
-      eprintf "Creating pipe: %a@." pp_print_pipe pipe;
+      eprintf "Creating process: %a@." pp_print_string_pipe pipe;
    match pipe with
       (*
        * The restriction to stdout and stderr is necessary to
@@ -1189,6 +1142,8 @@ let waitpid venv pos pid =
    let pos = string_pos "waitpid" pos in
       match pid with
          ExternalPid pid ->
+            if !debug_shell then
+               eprintf "Omake_shell_job.waitpid: external id %i@." pid;
             let _, status = Unix.waitpid [] pid in
                pid, status, ValNone
        | InternalPid pid ->
@@ -1199,12 +1154,9 @@ let waitpid venv pos pid =
        | ResultPid (code, value) ->
             0, Unix.WEXITED code, value
 
-(*!
- * @docoff
- *
+(*
  * -*-
  * Local Variables:
- * Caml-master: "compile"
  * End:
  * -*-
  *)
