@@ -8,25 +8,45 @@
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
- * as published by the Free Software Foundation; either version 2
- * of the License, or (at your option) any later version.
- *
+ * as published by the Free Software Foundation; version 2
+ * of the License.
+ * 
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
- *
+ * 
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
+ * 
+ * Additional permission is given to link this library with the
+ * with the Objective Caml runtime, and to redistribute the
+ * linked executables.  See the file LICENSE.OMake for more details.
  *
  * Author: Jason Hickey
  * @email{jyh@cs.caltech.edu}
  * @end[license]
  *)
-open Omake_shell_type
+open Lm_glob
+open Lm_symbol
+open Lm_location
 
-val parse : string -> pipe
+open Omake_env
+open Omake_shell_type
+open Omake_command_type
+
+(*
+ * Commands with a leading \ are quoted.
+ *)
+val parse_command_string : string -> simple_exe
+
+(*
+ * Construct the pipe from the value.
+ *)
+val pipe_of_value : venv ->
+   (venv -> pos -> loc -> string -> (symbol * apply) option) -> glob_options ->
+   pos -> loc -> value -> command_flag list * arg_pipe
 
 (*!
  * @docoff
